# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import
from .converter import ConverterFromNNEF as _Converter, Transform
from ..model import Tensor, Operation
from ..model.utils import generate_op_names_from_op_type
from ..utils import types
import numpy as np
import copy


class Converter(_Converter):

    @staticmethod
    def defined_shapes():
        return {
            'relu6': lambda shape: shape,
        }

    @staticmethod
    def decomposed_operations():
        return _Converter.decomposed_operations() + ['linear']

    def __init__(self, data_format='NXC', io_transpose=False, custom_transforms=None, custom_functions=None,
                 mirror_unsupported=False):
        _Converter.__init__(self, transforms=self.merge_transforms(_Transforms, custom_transforms),
                            functions=custom_functions, mirror_unsupported=mirror_unsupported)
        self._data_format = data_format
        self._io_transpose = io_transpose

    def __call__(self, graph):
        self.convert_variables_to_constants(graph)
        graph = _Converter.__call__(self, graph)
        self._fix_output_transposes(graph)
        self._remove_unused_constants(graph)
        generate_op_names_from_op_type(graph)
        return graph

    def _global_attribs(self):
        return {'_lite_': False}

    def _prepare(self, graph):
        self._fix_inline_constants(graph)

    def _fix_inline_constants(self, graph):
        for tensor in graph.tensors:
            mapped = self._tensor_map[tensor]
            if not mapped.producer and mapped.data is not None:
                self._const_operation(tensor, tensor.data)

    def _remove_unused_constants(self, graph):
        ops = [op for op in graph.operations if op.type == 'Const' and not op.output.has_consumer]
        tensors = [op.output for op in ops]
        graph.outputs = [tensor for tensor in graph.outputs if tensor not in tensors]
        graph.remove_operations(ops, unlink=True)
        graph.remove_tensors(tensors)

    def _fix_output_transposes(self, graph):
        graph.outputs = [self.transpose_input(tensor) if self.needs_io_transpose(tensor) else
                         self.undo_transpose(tensor) for tensor in graph.outputs]

    def _const_operation(self, output, value):
        Operation(output.graph, type='Const', inputs=(), outputs=output,
                  attribs={'value': types.to_numpy(value, dtype=output.dtype), 'dtype': output.dtype})

    def _transpose_operation(self, input, output, perm):
        Operation(input.graph, type='Transpose', inputs=(input, self.as_tensor(perm, np.int32)),
                  outputs=output, attribs={'T': input.dtype})

    def _reshape_operation(self, input, output, shape):
        Operation(input.graph, type='Reshape', inputs=(input, self.as_tensor(shape, np.int32)), outputs=output,
                  attribs={'T': input.dtype})

    def _squeeze_operation(self, input, output, axes):
        Operation(input.graph, type='Squeeze', inputs=input, outputs=output,
                  attribs={'squeeze_dims': axes, 'T': input.dtype})

    def _unsqueeze_operation(self, input, output, axes):
        if len(axes) == 1:
            Operation(input.graph, type='ExpandDims', inputs=(input, self.as_tensor(axes[0], np.int32)),
                      outputs=output, attribs={'T': input.dtype})
        else:
            Operation(input.graph, type='Reshape', inputs=(input, self.as_tensor(output.shape, np.int32)),
                      outputs=output, attribs={'T': input.dtype})

    def _scale_operation(self, input, output, scalar):
        if not isinstance(scalar, Tensor):
            scalar = self.as_tensor(scalar, np.float32)

        Operation(input.graph, type='Mul', inputs=(input, scalar), outputs=output,
                  attribs={'T': input.dtype})

    def _bias_operation(self, input, output, bias):
        if not isinstance(bias, Tensor):
            bias = self.as_tensor(bias, np.float32)

        if bias.rank == 1:
            Operation(output.graph, type='BiasAdd', inputs=(input, bias), outputs=output, attribs={'T': output.dtype})
        else:
            Operation(output.graph, type='Add', inputs=(input, bias), outputs=output, attribs={'T': output.dtype})

    def _make_constant(self, graph, dtype, value, inline):
        tensor = Tensor(graph, dtype=dtype, shape=self._shape_of(value))
        self._const_operation(tensor, value)
        return tensor

    def _transform_constant(self, tensor, func):
        data = func(tensor.producer.attribs['value'])
        tensor.shape = data.shape
        tensor.producer.attribs['value'] = data

    def _is_conv_filter(self, tensor, groups):
        tensor = self._tensor_map.get(tensor)
        return tensor and len(tensor.consumers) > 0 and \
               all(op.type == 'conv' and op.inputs[1] is tensor and op.attribs['groups'] == groups
                   for op in tensor.consumers)

    def _ensure_constant_producer(self, tensor):
        if tensor.is_constant and tensor.producer is None:
            Operation(tensor.graph, type='Const', inputs=(), outputs=tensor,
                      attribs={'value': tensor.data, 'dtype': tensor.data.dtype.type})

    def _is_nxc(self, format):
        return format[0] == 'N' and format[-1] == 'C' and len(format) > 2

    def _is_xcn(self, format):
        return format[-2] == 'C' and format[-1] == 'N' and len(format) > 2

    def _is_cxn(self, format):
        return format[0] == 'C' and format[-1] == 'N' and len(format) > 2

    def needs_io_transpose(self, tensor):
        if tensor.rank <= 2:
            return False
        if isinstance(self._io_transpose, bool):
            return self._io_transpose
        else:
            return tensor.name in self._io_transpose

    def is_nxc(self):
        return self._is_nxc(self._data_format)

    def data_format(self, rank):
        X = 'W' if rank == 1 else 'HW' if rank == 2 else 'DHW' if rank == 3 else None
        return self._data_format.replace('X', X) if X else self._data_format

    def convert_padding(self, value):
        return 'SAME' if value == [] else 'VALID' if all(item == (0, 0) for item in value) else 'EXPLICIT'

    def convert_explicit_paddings(self, value):
        if value == [] or all(item == (0, 0) for item in value):
            return None
        else:
            paddings = [item for pair in value for item in pair]
            return [0, 0] + paddings + [0, 0] if self.is_nxc() else [0, 0, 0, 0] + paddings

    def convert_size(self, value):
        if isinstance(value, tuple):
            return (1,) + value + (1,) if self.is_nxc() else (1, 1) + value[2:]
        else:
            return [1] + value + [1] if self.is_nxc() else [1, 1] + value[2:]

    def transpose_input(self, tensor):
        if self.is_nxc():
            return self._pre_transpose(tensor, self.ncx_to_nxc_perm(tensor.rank)) \
                if not self.transposing(tensor) and tensor.rank > 2 else tensor
        else:
            assert not self.transposing(tensor)
            return tensor

    def transpose_output(self, tensor):
        if self.is_nxc():
            self._transposes[tensor] = self.ncx_to_nxc(tensor.shape)
        return tensor

    def transpose_filter(self, tensor, format='XCN'):
        if self._is_xcn(format):
            perm = self.ncx_to_xcn_perm(tensor.rank)
        elif self._is_nxc(format):
            perm = self.ncx_to_nxc_perm(tensor.rank)
        elif self._is_cxn(format):
            perm = self.ncx_to_cxn_perm(tensor.rank)
        else:
            assert False

        return self._pre_transpose(tensor, perm)

    def transpose_depthwise_filter(self, tensor, channels, format='XCN'):
        if self._is_xcn(format):
            perm = self.ncx_to_xcn_perm(tensor.rank)
        elif self._is_nxc(format):
            perm = self.ncx_to_nxc_perm(tensor.rank)
        elif self._is_cxn(format):
            perm = self.ncx_to_cxn_perm(tensor.rank)
        else:
            assert False

        shape = tensor.shape[2:] + (channels, tensor.shape[0] // channels)
        return self._reshape(self._pre_transpose(tensor, perm), shape)

    def transpose_like(self, tensor, reference):
        if self.transposing(reference):
            self.transpose_output(tensor)
        return tensor

    def transpose_list_like(self, items, ref):
        return self.ncx_to_nxc(items) if self.transposing(ref) else items

    def transpose_axis_like(self, axis, ref, rank=None):
        return self.axis_ncx_to_nxc(axis, rank or ref.rank) if self.transposing(ref) else axis

    def undo_transpose(self, tensor):
        perm = self.nxc_to_ncx_perm(tensor.rank)
        if perm == list(range(tensor.rank)):
            return tensor
        return self._pre_transpose(tensor, perm) if self.transposing(tensor) else tensor

    def squeeze_input(self, tensor, axes):
        return self._pre_squeeze(tensor, axes=axes)

    def squeeze_output(self, tensor, axes):
        return self._post_squeeze(tensor, axes=axes)

    def unsqueeze_input(self, tensor, axes):
        return self._pre_unsqueeze(tensor, axes=axes)

    def unsqueeze_output(self, tensor, axes):
        return self._post_unsqueeze(tensor, axes=axes)

    def squeeze_vector(self, tensor):
        if self._is_constant(tensor) and len(self._tensor_map[tensor].consumers) == 1:
            self._transform_constant(tensor, lambda data: np.squeeze(data, 0))
            return tensor
        else:
            return self.squeeze_input(tensor, axes=[0])

    def scale_output(self, output, scalar):
        input = Tensor(output.graph, dtype=output.dtype, shape=self._working_shape(output), quant=copy.deepcopy(output.quant))
        self._scale_operation(input, output, scalar)
        return input

    def bias_add(self, output, bias):
        if bias.rank == 0 and np.all(bias.data == 0):
            return output

        input = Tensor(output.graph, dtype=output.dtype, shape=self._working_shape(output), quant=copy.deepcopy(output.quant))
        self._bias_operation(input, output, bias)
        return input

    def split_sizes(self, ratios, size):
        p = size / sum(ratios)
        return [p * r for r in ratios]

    def convert_binarg(self, tensor, other):
        self._ensure_constant_producer(tensor)
        if tensor.rank == 0:
            return tensor
        needs_transpose = self.transposing(other) and not self.transposing(tensor)
        if other.rank > tensor.rank:
            if tensor.rank == 2 and tensor.shape[0] == 1 and needs_transpose:
                return self.squeeze_vector(tensor)
            tensor = self._pre_unsqueeze(tensor, axes=list(range(tensor.rank, other.rank)))
        return self.transpose_input(tensor) if needs_transpose else tensor

    def as_numpy(self, value, dtype=None):
        return types.to_numpy(value, dtype)

    def as_bits(self, items):
        bits = 0
        for idx, val in enumerate(items):
            if val:
                bits |= (1 << idx)
        return bits

    def out_of_range(self, x, limit):
        return x >= limit or x <= -limit


_Transforms = Converter.unpack_transforms({
    'external':
        Transform(
            type='Placeholder',
            using={'needs_transpose': '!needs_io_transpose(O[0])'},
            outputs='!transpose_output(O[0]) if needs_transpose else O[0]',
            attribs={
                'shape': '!tuple(ncx_to_nxc(shape) if needs_transpose else shape)',
                'dtype': '!dtype',
            }
        ),
    'constant':
        Transform(
            type='Const',
            outputs='!O[0]',
            attribs={
                'dtype': '!O[0].dtype',
                'value': '!value if isinstance(value, np.ndarray) else as_numpy(value[0] if shape == [] else value)',
            }
        ),
    'conv':
        Transform(
            type='!"Conv{n}D".format(n=I[0].rank - 2) if groups != 0 else "DepthwiseConv2dNative"',
            cond={
                '!I[0].rank == 4 or I[0].rank == 5': 'rank must be 4 or 5',
            },
            using={
                'channels': '!I[0].shape[1]',
            },
            inputs=(
                '!transpose_input(I[0])',
                '!transpose_filter(I[1]) if groups != 0 else transpose_depthwise_filter(I[1], channels)',
            ),
            outputs=(
                '!bias_add(transpose_output(O[0]), squeeze_vector(I[2]) if I[2].rank == 2 else I[2])',
            ),
            attribs={
                'padding': '!convert_padding(padding)',
                'explicit_paddings': '!convert_explicit_paddings(padding)',
                'strides': '!convert_size(stride)',
                'dilations': '!convert_size(dilation)',
                'data_format': '!data_format(I[0].rank - 2)',
                'T': '!I[0].dtype',
            }
        ),
    'deconv':
        Transform(
            type='!"Conv{n}DBackpropInput".format(n=I[0].rank - 2) if groups != 0 else "DepthwiseConv2dNativeBackpropInput"',
            cond={
                '!I[0].rank == 4 or I[0].rank == 5': 'rank must be 4 or 5',
            },
            using={
                'channels': '!O[0].shape[1]',
            },
            inputs=(
                '!as_tensor(ncx_to_nxc(output_shape) if is_nxc() else output_shape, np.int32)',
                '!transpose_filter(I[1]) if groups != 0 else transpose_depthwise_filter(I[1], channels)',
                '!transpose_input(I[0])',
            ),
            outputs=(
                '!bias_add(transpose_output(O[0]), squeeze_vector(I[2]) if I[2].rank == 2 else I[2])',
            ),
            attribs={
                'padding': '!convert_padding(padding)',
                'explicit_paddings': '!convert_explicit_paddings(padding)',
                'strides': '!convert_size(stride)',
                'dilations': '!convert_size(dilation)',
                'data_format': '!data_format(I[0].rank - 2)',
                'T': '!I[0].dtype',
            }
        ),
    ('max_pool', 'avg_pool', 'max_pool_with_index'):
        Transform(
            type=('MaxPool', 'AvgPool', 'MaxPoolWithArgmax'),
            inputs=(
                '!transpose_input(I[0])',
            ),
            outputs=(
                '!transpose_output(O[0])',
                '!transpose_output(O[1]) if len(O) > 1 else None',
            ),
            attribs={
                'ksize': '!ncx_to_nxc(size) if is_nxc() else size',
                'strides': '!ncx_to_nxc(stride) if is_nxc() else stride',
                'padding': '!convert_padding(padding)',
                'explicit_paddings': '!convert_explicit_paddings(padding)',
                'data_format': '!data_format(I[0].rank - 2) if _type_ != "max_pool_with_index" else None',
                'T': '!I[0].dtype',
            }
        ),
    'box':
        Transform(
            type='AvgPool',
            using={'volume': '!int(np.prod(size))'},
            inputs=(
                '!transpose_input(I[0])',
            ),
            outputs=(
                '!scale_output(transpose_output(O[0]), volume) if not normalize else transpose_output(O[0])',
            ),
            attribs={
                'ksize': '!ncx_to_nxc(size) if is_nxc() else size',
                'strides': '!ncx_to_nxc(stride) if is_nxc() else stride',
                'padding': '!convert_padding(padding)',
                'explicit_paddings': '!convert_explicit_paddings(padding)',
                'data_format': '!data_format(I[0].rank - 2)',
                'T': '!I[0].dtype',
            }
        ),
    'reshape':
        Transform(
            type='Reshape',
            inputs=(
                '!undo_transpose(I[0])',
                '!as_tensor(fixed_batch(shape, I[0].shape[0]), np.int32)',
            ),
            outputs='!O[0]',
            attribs={
                'T': '!dtype if not _lite_ else None',
            }
        ),
    'transpose':
        Transform(
            type='Transpose',
            inputs=(
                '!I[0]',
                '!as_tensor(transpose_axis_like(axes, I[0]), np.int32)',
            ),
            outputs='!O[0]',
            attribs={
                'T': '!dtype if not _lite_ else None',
            }
        ),
    'squeeze':
        Transform(
            type='Squeeze',
            inputs='!undo_transpose(I[0])',
            outputs='!O[0]',
            attribs={
                'squeeze_dims': '!axes',
                'T': '!dtype if not _lite_ else None',
            }
        ),
    'unsqueeze':
        Transform(
            type='!"ExpandDims" if len(axes) == 1 else "Reshape"',
            using={
                'new_shape': '!unsqueeze_shape(I[0].shape, axes)',
            },
            inputs=(
                '!undo_transpose(I[0])',
                '!as_tensor(axes if len(axes) == 1 else new_shape, np.int32)',
            ),
            outputs='!O[0]',
            attribs={
                'T': '!dtype if not _lite_ else None',
                'new_shape': '!new_shape if _lite_ else None',
            }
        ),
    'stack':
        Transform(
            type='Pack',
            inputs=['![undo_transpose(t) for t in I]'],
            outputs='!O[0]',
            attribs={
                'axis': '!axis',
                'N': '!len(I) if not _lite_ else None',
                'values_count': '!len(I) if _lite_ else None',
                'T': '!dtype if not _lite_ else None',
            }
        ),
    'unstack':
        Transform(
            type='Unpack',
            inputs='!undo_transpose(I[0])',
            outputs=['!O[:]'],
            attribs={
                'axis': '!axis',
                'num': '!len(O)',
                'T': '!dtype if not _lite_ else None',
            }
        ),
    ('min_reduce', 'max_reduce', 'mean_reduce', 'sum_reduce', 'any_reduce', 'all_reduce'):
        Transform(
            type=('Min', 'Max', 'Mean', 'Sum', 'Any', 'All'),
            using={'dims': '!transpose_axis_like(axes, I[0])'},
            inputs=(
                '!I[0]',
                '!as_tensor(dims, np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'keep_dims': True,
                'T': '!I[0].dtype if I[0].dtype != bool and not _lite_ else None',
            }
        ),
    'concat':
        Transform(
            type='Concat',
            using={
                'dim': '!transpose_axis_like(axis, I[0])'
            },
            inputs=['!as_tensor(dim, np.int32)', '!I[:]'],
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'N': '!len(I)',
                'T': '!O[0].dtype if not _lite_ else None',
            }
        ),
    'split':
        Transform(
            type='SplitV',
            using={
                'dim': '!transpose_axis_like(axis, I[0])'
            },
            inputs=(
                '!I[0]',
                '!as_tensor(split_sizes(ratios, I[0].shape[axis]), np.int64)',
                '!as_tensor(dim, np.int32)',
            ),
            outputs=['![transpose_like(O[i], I[0]) for i in range(len(O))]'],
            attribs={
                'num_split': '!len(ratios) if not _lite_ else None',
                'num_splits': '!len(ratios) if _lite_ else None',
                'T': '!I[0].dtype if not _lite_ else None',
            }
        ),
    ('add', 'sub', 'mul', 'div', 'pow', 'lt', 'gt', 'le', 'ge', 'eq', 'ne', 'min', 'max', 'and', 'or'):
        Transform(
            type=('Add', 'Sub', 'Mul', 'RealDiv', 'Pow', 'Less', 'Greater', 'LessEqual', 'GreaterEqual',
                  'Equal', 'NotEqual', 'Minimum', 'Maximum', 'LogicalAnd', 'LogicalOr'),
            inputs=(
                '!convert_binarg(I[0], I[1])',
                '!convert_binarg(I[1], I[0])',
            ),
            outputs='!transpose_output(O[0]) if transposing(I[0]) or transposing(I[1]) else O[0]',
            attribs={
                'T': '!I[0].dtype if I[0].dtype != bool and not _lite_ else None',
            }
        ),
    ('copy', 'relu', 'relu6', 'elu', 'selu', 'gelu', 'silu', 'sigmoid', 'softplus', 'exp', 'log',
     'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
     'neg', 'rcp', 'sign', 'abs', 'floor', 'ceil', 'round', 'sqr', 'sqrt', 'rsqrt', 'not'):
        Transform(
            type=('Identity', 'Relu', 'Relu6', 'Elu', 'Selu', 'Gelu', 'Silu', 'Sigmoid', 'Softplus', 'Exp', 'Log',
                  'Sin', 'Cos', 'Tan', 'Asin', 'Acos', 'Atan', 'Sinh', 'Cosh', 'Tanh', 'Asinh', 'Acosh', 'Atanh',
                  'Neg', 'Reciprocal', 'Sign', 'Abs', 'Floor', 'Ceil', 'Round', 'Square', 'Sqrt', 'Rsqrt', 'LogicalNot'),
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'T': '!I[0].dtype if I[0].dtype != bool and not _lite_ else None',
            }
        ),
    'leaky_relu':
        Transform(
            type='LeakyRelu',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'alpha': '!alpha',
                'T': '!I[0].dtype if not _lite_ else None',
            }
        ),
    'batch_normalization':
        Transform(
            type='FusedBatchNorm',
            using={
                'channels': '!O[0].shape[1]'
            },
            inputs=(
                '!transpose_input(I[0])',
                '!squeeze_vector(I[4])',
                '!squeeze_vector(I[3])',
                '!squeeze_vector(I[1])',
                '!squeeze_vector(I[2])',
            ),
            outputs=(
                '!transpose_output(O[0])',
                '!new_tensor(shape=(channels,), dtype=O[0].dtype)',
                '!new_tensor(shape=(channels,), dtype=O[0].dtype)',
                '!new_tensor(shape=(channels,), dtype=O[0].dtype)',
                '!new_tensor(shape=(channels,), dtype=O[0].dtype)',
            ),
            attribs={
                'epsilon': '!epsilon',
                'data_format': '!data_format(I[0].rank - 2)',
                'T': '!I[0].dtype if not _lite_ else None',
                'is_training': False,
            }
        ),
    'softmax':
        Transform(
            type='Softmax',
            cond={
                '!axes == [1])': 'axes must equal channel dimension (1)',
            },
            inputs='!transpose_input(I[0])',
            outputs='!transpose_output(O[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'beta': '!1.0 if _lite_ else None',
            }
        ),
    'matmul':
        Transform(
            type='MatMul',
            inputs=('!I[0]', '!I[1]'),
            outputs='!O[0]',
            attribs={
                'transpose_a': '!transposeA',
                'transpose_b': '!transposeB',
                'T': '!I[0].dtype if not _lite_ else None',
            },
        ),
    'clamp':
        Transform(
            type='ClipByValue',
            inputs=('!I[0]', '!I[1]', '!I[2]'),
            outputs='!transpose_like(O[0], I[0])',
            attribs={'T': '!I[0].dtype if not _lite_ else None'},
        ),
    'pad':
        Transform(
            type='!"Pad" if border == "constant" else "MirrorPad"',
            cond={
                '!border in ["constant", "reflect", "reflect-even"]':
                    'border must be one of "constant", "reflect", "reflect-even"',
            },
            using={'paddings': '![list(item) for item in padding]'},
            inputs=(
                '!I[0]',
                '!as_tensor(ncx_to_nxc(paddings, cond=transposing(I[0])), np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'mode': '!"REFLECT" if border == "reflect" else "SYMMETRIC" if border == "reflect-even" else None',
            },
        ),
    'tile':
        Transform(
            type='Tile',
            inputs=(
                '!I[0]',
                '!as_tensor(transpose_list_like(repeats, I[0]), np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
            },
        ),
    'slice':
        Transform(
            type='StridedSlice',
            using={
                'dims': '!ncx_to_nxc(list(range(I[0].rank)), cond=transposing(I[0]))',
                'axis': '!ncx_to_nxc(axes, cond=transposing(I[0]))',
                'begs': '!ncx_to_nxc(begin, cond=transposing(I[0]))',
                'ends': '!ncx_to_nxc(end, cond=transposing(I[0]))',
                'strs': '!ncx_to_nxc(stride, cond=transposing(I[0]))',
            },
            inputs=(
                '!I[0]',
                '!as_tensor([begs[axis.index(i)] if i in axis else 0 for i in dims], np.int32)',
                '!as_tensor([ends[axis.index(i)] if i in axis else 0 for i in dims], np.int32)',
                '!as_tensor([strs[axis.index(i)] if i in axis else 1 for i in dims], np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'Index': '!np.int32',
                'begin_mask': '!as_bits([1 if i not in axis or out_of_range(begs[axis.index(i)], I[0].shape[i]) else 0'
                              ' for i in dims])',
                'end_mask': '!as_bits([1 if i not in axis or (ends[axis.index(i)] == 0 and strs[axis.index(i)] == 1)'
                            ' or out_of_range(ends[axis.index(i)], I[0].shape[i]) else 0 for i in dims])',
                'ellipsis_mask': 0,
                'new_axis_mask': 0,
                'shrink_axis_mask': 0,
            },
        ),
    ('argmin_reduce', 'argmax_reduce'):
        Transform(
            type=('ArgMin', 'ArgMax'),
            cond={
                '!len(axes) == 1': 'axes must be of length 1',
            },
            using={'axis': '!transpose_axis_like(axes[0], ref=I[0])'},
            inputs=(
                '!I[0]',
                '!as_tensor(axis, np.int32)',
            ),
            outputs='!transpose_like(unsqueeze_output(O[0], axes) if not _lite_ else O[0], I[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'output_type': '!O[0].dtype',
            }
        ),
    'select':
        Transform(
            type='Select',
            inputs=(
                '!I[0]',
                '!convert_binarg(I[1], I[0])',
                '!convert_binarg(I[2], I[0])',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'T': '!I[1].dtype if not _lite_ else None',
            }
        ),
    'nearest_upsample':
        Transform(
            type='ResizeNearestNeighbor',
            using={
                'size': '!I[0].shape[2:]'
            },
            inputs=(
                '!transpose_input(I[0])',
                '!as_tensor([s * f for s, f in zip(size, factor)], np.int32)',
            ),
            outputs='!transpose_output(O[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'align_corners': False,
                'half_pixel_centers': False,
            }
        ),
    'multilinear_upsample':
        Transform(
            type='ResizeBilinear',
            using={
                'size': '!I[0].shape[2:]'
            },
            inputs=(
                '!transpose_input(I[0])',
                '!as_tensor([s * f for s, f in zip(size, factor)], np.int32)',
            ),
            outputs='!transpose_output(O[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'align_corners': '!method == "aligned"',
                'half_pixel_centers': '!method == "symmetric"',
            }
        ),
    ('nearest_downsample', 'area_downsample'):
        Transform(
            type=('ResizeNearestNeighbor', 'ResizeArea'),
            using={'size': '!I[0].shape[2:]'},
            inputs=(
                '!transpose_input(I[0])',
                '!as_tensor([s // f for s, f in zip(size, factor)], np.int32)',
            ),
            outputs='!transpose_output(O[0])',
            attribs={
                'T': '!I[0].dtype if not _lite_ else None',
                'align_corners': False,
                'half_pixel_centers': '!False if _type_ == "nearest_downsample" else None',
            }
        ),
    'local_response_normalization':
        Transform(
            type='LRN',
            cond={
                '!all(s == 1 or i == 1 for i, s in enumerate(size))':
                    'size must be singular for all non-channel dimensions',
            },
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'depth_radius': '!size[1] // 2 if not _lite_ else None',
                'radius': '!size[1] // 2 if _lite_ else None',
                'alpha': '!alpha / size[1]',
                'beta': '!beta',
                'bias': '!bias',
            }
        ),
    'add_n':
        Transform(
            type='AddN',
            inputs=['!I[:]'],
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'T': '!O[0].dtype',
                'N': '!len(I)',
            },
        ),
    'cast':
        Transform(
            type='Cast',
            inputs='!I[0]',
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'SrcT': '!I[0].dtype',
                'DstT': '!O[0].dtype',
            },
        ),
    'gather':
        Transform(
            type='GatherV2',
            inputs=(
                '!I[0]',
                '!I[1]',
                '!as_tensor(transpose_axis_like(axis, I[0]), np.int32)',
            ),
            outputs='!transpose_like(O[0], I[0])',
            attribs={
                'Tparams': '!I[0].dtype',
                'Tindices': '!I[1].dtype',
                'Taxis': np.int32,
            },
        ),
})



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class DimensionMetadata(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDimensionMetadata(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DimensionMetadata()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def DimensionMetadataBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x54\x46\x4C\x33", size_prefixed=size_prefixed)

    # DimensionMetadata
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DimensionMetadata
    def Format(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # DimensionMetadata
    def DenseSize(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # DimensionMetadata
    def ArraySegmentsType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # DimensionMetadata
    def ArraySegments(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # DimensionMetadata
    def ArrayIndicesType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # DimensionMetadata
    def ArrayIndices(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

def DimensionMetadataStart(builder): builder.StartObject(6)
def DimensionMetadataAddFormat(builder, format): builder.PrependInt8Slot(0, format, 0)
def DimensionMetadataAddDenseSize(builder, denseSize): builder.PrependInt32Slot(1, denseSize, 0)
def DimensionMetadataAddArraySegmentsType(builder, arraySegmentsType): builder.PrependUint8Slot(2, arraySegmentsType, 0)
def DimensionMetadataAddArraySegments(builder, arraySegments): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(arraySegments), 0)
def DimensionMetadataAddArrayIndicesType(builder, arrayIndicesType): builder.PrependUint8Slot(4, arrayIndicesType, 0)
def DimensionMetadataAddArrayIndices(builder, arrayIndices): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(arrayIndices), 0)
def DimensionMetadataEnd(builder): return builder.EndObject()



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import

from ..model import *
from ..utils import types
import numpy as np
import functools
import inspect
import math
import copy
import six
import re


class Transform:

    def __init__(self, type, name=None, inputs=None, outputs=None, attribs=None,
                 defaults=None, using=None, cond=None, custom=False):
        self.type = type
        self.name = name or '!_name_'
        self.inputs = inputs or ()
        self.outputs = outputs or ()
        self.attribs = attribs or {}
        self.defaults = defaults
        self.using = using or {}
        self.cond = cond
        self.custom = custom

    def with_type(self, type):
        return Transform(type=type, name=self.name, inputs=self.inputs, outputs=self.outputs, attribs=self.attribs,
                         defaults=self.defaults, using=self.using, cond=self.cond, custom=self.custom)


class ConversionError(Exception):

    def __init__(self, message, details=None):
        Exception.__init__(self, message)
        self.details = details


class Converter:

    @staticmethod
    def find_public_methods(obj):
        methods = inspect.getmembers(obj, predicate=inspect.ismethod)
        return {name: func for name, func in methods if not name.startswith('_')}

    @staticmethod
    def find_public_functions(obj):
        methods = inspect.getmembers(obj, predicate=inspect.isfunction)
        return {name: func for name, func in methods if not name.startswith('_')}

    @staticmethod
    def decomposed_operations():
        return []       # return list of decomposed NNEF ops in subclass if converting from NNEF

    @staticmethod
    def defined_operations():
        return {}       # return dictionary of NNEF operator (fragment) definitions in subclass if converting to NNEF

    @staticmethod
    def defined_operation_dependencies():
        return {}  # return dictionary of NNEF operator (fragment) dependencies in subclass if converting to NNEF

    @staticmethod
    def defined_shapes():
        return {}       # return dictionary of shape functions for NNEF fragments defined by the converter

    @staticmethod
    def unpack_transforms(transforms):
        unpacked = {}
        for key, value in six.iteritems(transforms):
            assert isinstance(value, Transform)

            if isinstance(key, tuple):
                if isinstance(value, Transform) and isinstance(value.type, tuple):
                    for key_item, type_item in zip(key, value.type):
                        value_item = copy.deepcopy(value)
                        value_item.type = type_item
                        unpacked[key_item] = value_item
                else:
                    for item in key:
                        unpacked[item] = value
            else:
                unpacked[key] = value

        return unpacked

    @staticmethod
    def merge_transforms(default_transforms, custom_transforms):
        if custom_transforms is None:
            return default_transforms

        transforms = dict(default_transforms)
        transforms.update(custom_transforms)
        return transforms

    def __init__(self, transforms, functions=None, mirror_unsupported=False, infer_shapes=False, custom_shapes=None):
        self._graph = None
        self._transforms = transforms
        self._callables = self.find_public_methods(self)
        if functions:
            self._callables.update({name: functools.partial(func, self) for name, func in six.iteritems(functions)})
        self._mirror_unsupported = mirror_unsupported
        self._infer_shapes = infer_shapes
        self._custom_shapes = custom_shapes or {}

    def __call__(self, graph):
        if not self._infer_shapes:
            unknown_tensors = [tensor for tensor in graph.tensors if (tensor.shape is None or any(s is None for s in tensor.shape))
                               and len(tensor.consumers)]
            if len(unknown_tensors):
                names = ["'{}'".format(tensor.name) for tensor in unknown_tensors if tensor.name]
                raise ConversionError(("Input graph contains tensors with dynamic shape: " +
                                      ", ".join(names) if len(names) else "(no names)") +
                                      "\nTry the --fold-constants option to evaluate constant sub-graphs "
                                      "or the --static-only option to convert only the static part of the graph")

        self._graph = Graph(name=graph.name)
        self._tensor_map = {tensor: self._copy_tensor_(tensor) for tensor in graph.tensors}
        self._tensor_map.update({val: key for key, val in six.iteritems(self._tensor_map)})
        self._transposes = {}

        self._prepare(self._graph)

        if not self._infer_shapes:
            errors = []
            for op in graph.operations:
                transform = self._transforms.get(op.type)
                if transform is None and self._mirror_unsupported:
                    continue

                if isinstance(transform, Transform) and transform.type is None:
                    continue

                error = self._error_message(op, transform)
                if error is not None:
                    errors.append(error)

            if len(errors):
                raise ConversionError("Found {} operator(s) that cannot be converted\n{}".format(len(errors), "\n".join(errors)))

        for op in graph.operations:
            transform = self._transforms.get(op.type)
            if isinstance(transform, Transform) and transform.type is None:
                continue

            if self._infer_shapes:
                error = self._error_message(op, transform)
                if error is not None:
                    raise ConversionError(error)

            count = len(self._graph.operations)

            if transform is not None:
                self._convert(op, transform)
            elif self._mirror_unsupported:
                self._mirror(op)

            if self._infer_shapes:
                from nnef.shapes import _infer_op_shapes
                for op in self._graph.operations[count:]:
                    input_shapes = [list(tensor.shape) for tensor in op.inputs]
                    if not isinstance(op.inputs, tuple):
                        input_shapes = (input_shapes,)

                    output_counts = [len(op.outputs)] if not isinstance(op.outputs, tuple) else [None] * len(op.outputs)
                    output_shapes = _infer_op_shapes(op.type, op.attribs, input_shapes, output_counts,
                                                     custom_shapes=self._custom_shapes)
                    if not isinstance(op.outputs, tuple):
                        output_shapes = output_shapes[0]

                    for output, shape in zip(op.outputs, output_shapes):
                        output.shape = tuple(shape)

        for tensor, shape in self._transposes.items():
            tensor.shape = shape

        self._graph.remove_tensors([tensor for tensor in self._graph.tensors
                                    if len(tensor.producers) == 0 and len(tensor.consumers) == 0])

        self._graph.inputs = tuple(self._tensor_map[tensor] for tensor in graph.inputs if self._tensor_map[tensor].graph)
        self._graph.outputs = tuple(self._tensor_map[tensor] for tensor in graph.outputs if self._tensor_map[tensor].graph)

        return self._graph

    def tensor_mapping(self):
        return {key.name: value.name for key, value in six.iteritems(self._tensor_map)
                if value.graph == self._graph and key.name is not None and value.name is not None}

    def _global_attribs(self):
        return {}

    def _prepare(self, graph):
        pass

    def _check_conditions(self, op, transform):
        op_inputs = list(self._tensor_map[tensor] for tensor in op.inputs)
        op_outputs = list(self._tensor_map[tensor] for tensor in op.outputs)
        op_attribs = self._add_default_attribs(op.attribs, transform.defaults, op_inputs, op_outputs, op.type, op.name)

        using = {'_type_': op.type, '_name_': op.name, **self._global_attribs()}
        for key, item in six.iteritems(transform.using):
            value = self._evaluate(op_attribs, op_inputs, op_outputs, item, using)
            self._check_value(value, 'using', key, op.type, op.name)
            using[key] = value

        error = None
        if transform.cond is not None:
            for condition, message in transform.cond.items():
                if not self._evaluate(op_attribs, op_inputs, op_outputs, condition, using):
                    error = message if error is None else error + ', ' + message

        return error

    def _error_message(self, op, transform):
        if transform is None:
            return "Conversion of operator '{}' is not implemented".format(op.type)

        message = self._check_conditions(op, transform)
        if message is not None:
            attribs = {key: value for key, value in six.iteritems(op.attribs) if not key.startswith('_')}
            input_shapes = ", ".join(str(tensor.shape) for tensor in op.inputs)
            output_shapes = ", ".join(str(tensor.shape) for tensor in op.outputs)
            return "Conversion of operator '{}' is not possible: {}"\
                   "\n  attributes: {}\n  input-shapes: {}\n  output-shapes: {}"\
                .format(op.type, message, attribs, input_shapes, output_shapes)

        return None

    def _convert(self, op, transform):
        op_inputs = list(self._tensor_map[tensor] for tensor in op.inputs)
        op_outputs = list(self._tensor_map[tensor] for tensor in op.outputs)
        op_attribs = self._add_default_attribs(op.attribs, transform.defaults, op_inputs, op_outputs, op.type, op.name)

        using = {'_type_': op.type, '_name_': op.name, **self._global_attribs()}
        for key, item in six.iteritems(transform.using):
            value = self._evaluate(op_attribs, op_inputs, op_outputs, item, using)
            self._check_value(value, 'using', key, op.type, op.name)
            using[key] = value

        type = self._evaluate(op_attribs, op_inputs, op_outputs, transform.type, using)
        self._check_value(type, 'field', 'type', op.type, op.name)

        name = self._evaluate(op_attribs, op_inputs, op_outputs, transform.name, using)
        self._check_value(name, 'field', 'name', op.type, op.name)

        attribs = {}
        for key, item in six.iteritems(transform.attribs):
            value = self._evaluate(op_attribs, op_inputs, op_outputs, item, using)
            if value is not None:
                attribs[key] = value

        for key, value in six.iteritems(attribs):
            self._check_value(value, 'attribute', key, op.type, op.name)

        if isinstance(transform.inputs, list):
            inputs = self._evaluate_tensor_list(op_attribs, op_inputs, op_outputs, transform.inputs, using)
        elif isinstance(transform.inputs, tuple):
            inputs = tuple(self._filter_none(self._evaluate(op_attribs, op_inputs, op_outputs, item, using)
                                             for item in transform.inputs))
        else:
            inputs = (self._evaluate(op_attribs, op_inputs, op_outputs, transform.inputs, using),)

        for idx, item in enumerate(inputs):
            self._check_value(item, 'input', idx, op.type, op.name, tensor=True)

        offset = len(self._graph.operations)

        if isinstance(transform.outputs, list):
            outputs = self._evaluate_tensor_list(op_attribs, op_inputs, op_outputs, transform.outputs, using)
        elif isinstance(transform.outputs, tuple):
            outputs = tuple(self._filter_none(self._evaluate(op_attribs, op_inputs, op_outputs, item, using)
                                              for item in transform.outputs))
        else:
            outputs = (self._evaluate(op_attribs, op_inputs, op_outputs, transform.outputs, using),)

        for idx, item in enumerate(outputs):
            self._check_value(item, 'output', idx, op.type, op.name, tensor=True)

        op = Operation(self._graph, type=type, name=name, attribs=attribs, inputs=inputs, outputs=outputs,
                       custom=transform.custom)
        self._graph.reverse(offset)
        return op

    def _mirror(self, op):
        inputs_type = tuple if isinstance(op.inputs, tuple) else list
        outputs_type = tuple if isinstance(op.outputs, tuple) else list

        op_inputs = inputs_type(self._tensor_map[tensor] for tensor in op.inputs)
        op_outputs = outputs_type(self._tensor_map[tensor] for tensor in op.outputs)

        return Operation(self._graph, type=op.type, name=op.name, attribs=op.attribs,
                         inputs=op_inputs, outputs=op_outputs, custom=True)

    def _add_default_attribs(self, attribs, defaults, inputs, outputs, op_type, op_name):
        if defaults is None:
            return attribs

        attribs = dict(attribs)
        for key, value in six.iteritems(defaults):
            if key not in attribs:
                value = self._evaluate({}, inputs, outputs, value)
                self._check_value(value, 'default', key, op_type, op_name)
                attribs[key] = value

        return attribs

    def _evaluate(self, attribs, inputs, outputs, arg, using={}):
        if isinstance(arg, str) and arg[0] == '!':
            try:
                return eval(arg[1:], {'I': inputs, 'O': outputs, **attribs, **using, **self._callables,
                                      'np': np, 'math': math})
            except Exception as e:
                return e
        else:
            return arg

    def _evaluate_tensor_list(self, attribs, inputs, outputs, arg, using):
        values = []
        for item in arg:
            value = self._evaluate(attribs, inputs, outputs, item, using)
            if isinstance(value, Tensor) or isinstance(value, Exception):
                values.append(value)
            else:
                assert isinstance(value, (list, tuple))
                values += list(value)
        return values

    def _filter_none(self, items):
        return (item for item in items if item is not None)

    def _check_value(self, value, kind, key, op_type, op_name, tensor=False):
        if isinstance(value, Exception):
            raise ConversionError("Could not evaluate {kind} '{key}' while converting operator '{type}'; {err}: {cause}"
                                  .format(kind=kind, key=key, type=op_type, name=op_name,
                                          err=type(value).__name__, cause=str(value) or repr(value)))
        if tensor and not isinstance(value, Tensor):
            raise ConversionError("While converting operator '{op_type}', {kind} '{key}' must result in a tensor, "
                                  "but found {value_type}"
                                  .format(kind=kind, key=key, op_type=op_type, value_type=type(value)))

    def _copy_tensor_(self, tensor):
        return Tensor(self._graph, name=tensor.name, dtype=tensor.dtype, shape=tensor.shape,
                      data=tensor.data, quant=copy.deepcopy(tensor.quant))

    def _read_constant(self, tensor, type):
        raise NotImplementedError()

    def _make_constant(self, graph, dtype, value, inline):
        raise NotImplementedError()

    def _const_operation(self, output, value):
        raise NotImplementedError()

    def _transpose_operation(self, input, output, perm):
        raise NotImplementedError()

    def _reshape_operation(self, input, output, shape):
        raise NotImplementedError()

    def _squeeze_operation(self, input, output, axes):
        raise NotImplementedError()

    def _unsqueeze_operation(self, input, output, axes):
        raise NotImplementedError()

    def _scale_operation(self, input, output, scalar):
        raise NotImplementedError()

    @staticmethod
    def _permute(items, perm):
        permuted = list(items)
        for i in range(len(perm)):
            permuted[i] = items[perm[i]]
        return type(items)(permuted)

    @staticmethod
    def _inverse_permute(items, perm):
        permuted = list(items)
        for i in range(len(perm)):
            permuted[perm[i]] = items[i]
        return type(items)(permuted)

    def _working_shape(self, tensor):
        return self._transposes.get(tensor) or tensor.shape

    def _pre_transpose(self, input, perm):
        shape = self._permute(self._working_shape(input), perm)
        output = Tensor(input.graph, dtype=input.dtype, shape=shape, quant=copy.deepcopy(input.quant))
        self._transpose_operation(input, output, perm)
        return output

    def _post_transpose(self, output, perm):
        shape = self._inverse_permute(self._working_shape(output), perm)
        input = Tensor(output.graph, dtype=output.dtype, shape=shape, quant=copy.deepcopy(output.quant))
        self._transpose_operation(input, output, perm)
        return input

    def _pre_squeeze(self, input, axes):
        shape = self.squeeze_shape(self._working_shape(input), axes)
        output = Tensor(input.graph, dtype=input.dtype, shape=shape, quant=copy.deepcopy(input.quant))
        self._squeeze_operation(input, output, axes)
        return output

    def _pre_unsqueeze(self, input, axes):
        shape = self.unsqueeze_shape(self._working_shape(input), axes)
        output = Tensor(input.graph, dtype=input.dtype, shape=shape, quant=copy.deepcopy(input.quant))
        self._unsqueeze_operation(input, output, axes)
        return output

    def _post_squeeze(self, output, axes):
        shape = self.unsqueeze_shape(self._working_shape(output), axes)
        input = Tensor(output.graph, dtype=output.dtype, shape=shape, quant=copy.deepcopy(output.quant))
        self._squeeze_operation(input, output, axes)
        return input

    def _post_unsqueeze(self, output, axes):
        shape = self.squeeze_shape(self._working_shape(output), axes)
        input = Tensor(output.graph, dtype=output.dtype, shape=shape, quant=copy.deepcopy(output.quant))
        self._unsqueeze_operation(input, output, axes)
        return input

    def _reshape(self, input, shape):
        output = Tensor(input.graph, dtype=input.dtype, shape=shape, quant=copy.deepcopy(input.quant))
        self._reshape_operation(input, output, shape)
        return output

    def _shape_of(self, value):
        if isinstance(value, (list, tuple)):
            length = len(value)
            return (length,) + self._shape_of(value[0]) if length > 0 else (0,)
        elif isinstance(value, np.ndarray):
            return value.shape
        else:
            return ()

    def squeeze_shape(self, shape, axes):
        return type(shape)(shape[i] for i in range(len(shape)) if i not in axes)

    def unsqueeze_shape(self, shape, axes):
        for axis in axes:
            shape = shape[:axis] + (1,) + shape[axis:]
        return shape

    def transposing(self, tensor):
        return tensor in self._transposes

    def nxc_to_ncx(self, items, cond=True):
        return items[0:1] + items[-1:] + items[1:-1] if cond else items

    def ncx_to_nxc(self, items, cond=True):
        return items[0:1] + items[2:] + items[1:2] if cond else items

    def xcn_to_ncx(self, items, cond=True):
        return items[-1:] + items[-2:-1] + items[:-2] if cond else items

    def ncx_to_xcn(self, items, cond=True):
        return items[2:] + items[1:2] + items[0:1] if cond else items

    def cxn_to_ncx(self, items, cond=True):
        return items[-1:] + items[:-1] if cond else items

    def ncx_to_cxn(self, items, cond=True):
        return items[1:] + items[:1] if cond else items

    def nxc_to_ncx_perm(self, rank):
        return self.nxc_to_ncx(list(range(rank)))

    def ncx_to_nxc_perm(self, rank):
        return self.ncx_to_nxc(list(range(rank)))

    def xcn_to_ncx_perm(self, rank):
        return self.xcn_to_ncx(list(range(rank)))

    def ncx_to_xcn_perm(self, rank):
        return self.ncx_to_xcn(list(range(rank)))

    def cxn_to_ncx_perm(self, rank):
        return self.cxn_to_ncx(list(range(rank)))

    def ncx_to_cxn_perm(self, rank):
        return self.ncx_to_cxn(list(range(rank)))

    def axis_nxc_to_ncx(self, value, rank):
        if isinstance(value, (list, tuple)):
            return type(value)(self.axis_nxc_to_ncx(v, rank) for v in value)
        else:
            if value < 0:
                value += rank
            return 0 if value == 0 else 1 if value == rank - 1 else value + 1

    def axis_ncx_to_nxc(self, value, rank):
        if isinstance(value, (list, tuple)):
            return type(value)(self.axis_ncx_to_nxc(v, rank) for v in value)
        else:
            if value < 0:
                value += rank
            return 0 if value == 0 else rank - 1 if value == 1 else value - 1

    def ensure_positive(self, axis, rank):
        if isinstance(axis, (list, tuple)):
            return type(axis)(self.ensure_positive(item, rank) for item in axis)
        else:
            return axis + rank if axis < 0 else axis

    def as_const(self, tensor, type=None):
        return self._read_constant(self._tensor_map[tensor], type=type)

    def is_const(self, tensor, type=None):
        tensor = self._tensor_map[tensor]
        if tensor.data is not None:
            return True
        try:
            self._read_constant(tensor, type=type)
            return True
        except ConversionError:
            return False

    def is_zero(self, tensor):
        return self.is_const(tensor) and len(tensor.shape) == 0 and self.as_const(tensor) == 0

    def as_tensor(self, arg, dtype, inline=None):
        return self._make_constant(self._graph, dtype=dtype, value=arg, inline=inline)

    def new_tensor(self, shape, dtype):
        return Tensor(self._graph, dtype=dtype, shape=shape)

    def is_integer_upsample(self, input_shape, output_shape):
        return all(output % input == 0 for input, output in zip(input_shape, output_shape))

    def is_integer_downsample(self, input_shape, output_shape):
        return all(input % output == 0 for input, output in zip(input_shape, output_shape))

    def upsample_factor(self, input_shape, output_shape):
        return [output // input for input, output in zip(input_shape, output_shape)]

    def downsample_factor(self, input_shape, output_shape):
        return [input // output for input, output in zip(input_shape, output_shape)]

    def from_numpy(self, array, type=None):
        return types.from_numpy(array, type)

    def to_numpy(self, value, dtype=None):
        return types.to_numpy(value, dtype)

    def flexible_batch(self, output_shape, batch):
        return [0] + output_shape[1:] if output_shape[0] == batch else output_shape

    def fixed_batch(self, output_shape, batch):
        return [batch] + output_shape[1:] if output_shape[0] == 0 else output_shape


class ConverterToNNEF(Converter):

    _DtypeFromNumpy = {
        np.float16: 'scalar',
        np.float32: 'scalar',
        np.float64: 'scalar',
        np.int8: 'integer',
        np.uint8: 'integer',
        np.int16: 'integer',
        np.uint16: 'integer',
        np.int32: 'integer',
        np.uint32: 'integer',
        np.int64: 'integer',
        np.uint64: 'integer',
        np.bool_: 'logical',
    }

    def __init__(self, transforms, functions=None, mirror_unsupported=False, infer_shapes=False, custom_shapes=None):
        Converter.__init__(self, transforms, functions, mirror_unsupported, infer_shapes, custom_shapes)

    def _insert_externals_and_constants(self, graph):
        for tensor in graph.tensors:
            mapped = self._tensor_map[tensor]
            if mapped.producer is None and len(mapped.consumers) > 0:
                if mapped.data is None:
                    Operation(graph, type='external', inputs=(), outputs=tensor,
                              attribs={'shape': list(tensor.shape), 'dtype': tensor.dtype})
                else:
                    Operation(graph, type='constant', inputs=(), outputs=tensor,
                              attribs={'shape': list(tensor.shape), 'dtype': tensor.dtype, 'value': mapped.data})

    def _ensure_valid_ids(self, graph):
        if graph.name is not None:
            graph.name = self.ensure_valid_id(graph.name)

        for tensor in graph.tensors:
            if tensor.name is not None:
                tensor.name = self.ensure_valid_id(tensor.name)

    def _make_constant(self, graph, dtype, value, inline):
        if isinstance(value, tuple):
            value = list(value)
        shape = value.shape if isinstance(value, np.ndarray) else (len(value),) if isinstance(value, list) else ()
        isarray = isinstance(value, np.ndarray) or isinstance(value, list)

        tensor = Tensor(graph, dtype=dtype, shape=shape)
        if inline:
            tensor.data = types.to_numpy(value, dtype)
        else:
            self._const_operation(tensor, value=value if isarray else [value])
        return tensor

    def _const_operation(self, output, value):
        Operation(output.graph, type='constant', inputs=(), outputs=output,
                  attribs={'value': value, 'dtype': output.dtype, 'shape': list(output.shape)})

    def _transpose_operation(self, input, output, perm):
        Operation(input.graph, type='transpose', inputs=input, outputs=output, attribs={'axes': perm})

    def _reshape_operation(self, input, output, shape):
        Operation(input.graph, type='reshape', inputs=input, outputs=output, attribs={'shape': list(shape)})

    def _squeeze_operation(self, input, output, axes):
        Operation(input.graph, type='squeeze', inputs=input, outputs=output, attribs={'axes': axes})

    def _unsqueeze_operation(self, input, output, axes):
        Operation(input.graph, type='unsqueeze', inputs=input, outputs=output, attribs={'axes': axes})

    def _scale_operation(self, input, output, scalar):
        if not isinstance(scalar, Tensor):
            scalar = self.as_tensor(scalar, np.float32)

        Operation(input.graph, type='mul', inputs=(input, scalar), outputs=output)

    def _bias_operation(self, input, output, bias):
        if not isinstance(bias, Tensor):
            bias = self.as_tensor(bias, np.float32)

        Operation(input.graph, type='add', inputs=(input, bias), outputs=output)

    def _transform_constant(self, tensor, func):
        data = func(tensor.producer.attribs['value'])
        tensor.shape = data.shape
        tensor.producer.attribs['value'] = data
        tensor.producer.attribs['shape'] = list(data.shape)

    @staticmethod
    def remove_unused_constants(graph):
        ops = [op for op in graph.operations if op.type == 'constant' and not op.output.has_consumer]
        tensors = [op.output for op in ops]
        graph.outputs = [tensor for tensor in graph.outputs if tensor not in tensors]
        graph.remove_operations(ops, unlink=True)
        graph.remove_tensors(tensors)

    @staticmethod
    def inline_scalar_constants(graph):
        for op in graph.operations:
            if op.type == 'constant':
                value = op.attribs['value']
                if not isinstance(value, np.ndarray):
                    value = np.array(value, op.output.dtype).reshape(op.output.shape)
                if len(value.shape) == 0:
                    op.output.data = value
                    graph.remove_operation(op, unlink=True)

    @staticmethod
    def convert_constants_to_variables(graph):
        variables = 0
        for op in graph.operations:
            if op.type == 'constant':
                value = op.attribs['value']
                if isinstance(value, np.ndarray):
                    variables += 1
                    op.type = 'variable'
                    op.attribs['label'] = op.name if op.name else 'variable' + str(variables)
                    op.output.data = value
                    del op.attribs['value']

    @staticmethod
    def ensure_valid_id(name):
        return re.sub('[^_0-9a-zA-Z]+', '_', name)

    def nnef_dtype(self, dtype):
        return ConverterToNNEF._DtypeFromNumpy[dtype]


class ConverterFromNNEF(Converter):

    @staticmethod
    def decomposed_operations():
        return ['separable_conv', 'separable_deconv', 'rms_pool',
                'local_mean_normalization', 'local_variance_normalization', 'local_contrast_normalization',
                'l1_normalization', 'moments']

    def __init__(self, transforms, functions=None, mirror_unsupported=False):
        Converter.__init__(self, transforms, functions, mirror_unsupported)

    @staticmethod
    def convert_variables_to_constants(graph):
        for op in graph.operations:
            if op.type == 'variable':
                op.type = 'constant'
                op.attribs['value'] = op.output.data
                del op.attribs['label']

    @staticmethod
    def fill_data_in_constants(graph):
        for op in graph.operations:
            if op.type == 'constant':
                op.output.data = op.attribs['value']

    def _is_constant(self, tensor):
        if tensor.producer:
            return tensor.producer.type == 'constant'
        else:
            return tensor.data is not None

    def _read_constant(self, tensor, type):
        if tensor.data is not None:
            value = tensor.data
        elif tensor.producer and tensor.producer.type == 'constant':
            value = tensor.producer.attribs['value']
        else:
            raise ConversionError('trying to evaluate non-constant tensor')

        return types.from_numpy(value, type=type) if isinstance(value, np.ndarray) else \
            types.cast(value, type=type) if type is not None else value



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .utils import stdio
from .interpreter import Statistics
from collections import namedtuple
import importlib
import argparse
import numpy as np
import json
import nnef
import six
import sys
import os


_onnx_dtype_to_numpy = {
    "tensor(float)": np.float32,
    "tensor(double)": np.float64,
    "tensor(int8)": np.int8,
    "tensor(int16)": np.int16,
    "tensor(int32)": np.int32,
    "tensor(int64)": np.int64,
    "tensor(uint8)": np.uint8,
    "tensor(uint16)": np.uint16,
    "tensor(uint32)": np.uint32,
    "tensor(uint64)": np.uint64,
    "tensor(bool)": np.bool_,
}

_nnef_dtype_to_numpy = {
    'scalar': np.float32,
    'integer': np.int32,
    'logical': np.bool_,
}

_numpy_dtype_remap = {
    np.short: np.int64,
    np.int_: np.int64,
    np.longlong: np.int64,
    np.ushort: np.uint64,
    np.uint: np.uint64,
    np.ulonglong: np.uint64,
    np.float_: np.float32,
    np.double: np.float64,
    np.longdouble: np.float64,
}


def _is_lambda(v):
    LAMBDA = lambda: 0
    return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__


def uniform(min=0, max=1):
    return lambda shape: np.random.uniform(min, max, shape)


def normal(mean=0, std=1):
    return lambda shape: np.random.normal(mean, std, shape)


def needs_transpose(io_transpose, name):
    return io_transpose is not None and (len(io_transpose) == 0 or name in io_transpose)


def transpose_channels_last_to_first(x):
    rank = len(x.shape)
    return np.transpose(x, axes=[0, rank - 1] + list(range(1, rank - 1)))


def transpose_channels_first_to_last(x):
    rank = len(x.shape)
    return np.transpose(x, axes=[0] + list(range(2, rank)) + [1])


def read_input(file, name, shape, dtype, transpose):
    data = nnef.read_tensor(file)

    any_batch = shape[0] == 0
    offset = int(any_batch)
    if tuple(data.shape[offset:]) != tuple(shape[offset:]):
        raise ValueError("Mismatch between declared and read shape for input '{}'; {} vs {}"
                         .format(name, data.shape, shape))
    if data.dtype != dtype:
        raise ValueError("Mismatch between declared and read dtype for input '{}'; {} vs {}"
                         .format(name, data.dtype, dtype))

    return transpose_channels_first_to_last(data) if transpose else data


def compute_statistics(array):
    if array.size == 0:
        return Statistics(num=0, min=0.0, max=0.0, sum=0.0, ssum=0.0)

    return Statistics(
        num=array.size,
        min=float(np.min(array)),
        max=float(np.max(array)),
        sum=float(np.sum(array)),
        ssum=float(np.sum(array * array)),
    )


class RandomInputSource:

    def __init__(self, distribution):
        self._distribution = distribution

    def __call__(self, name, shape, dtype):
        return self._distribution(shape).astype(dtype)


class StreamInputSource:

    def __init__(self, stream, io_transpose):
        self._stream = stream
        self._io_transpose = io_transpose

    def __call__(self, name, shape, dtype):
        return read_input(self._stream, name, shape, dtype,
                          transpose=needs_transpose(self._io_transpose, name))


class FileInputSource:

    def __init__(self, folder, io_transpose):
        self._folder = folder
        self._io_transpose = io_transpose

    def __call__(self, name, shape, dtype):
        with open(os.path.join(self._folder, name + '.dat')) as file:
            return read_input(file, name, shape, dtype,
                              transpose=needs_transpose(self._io_transpose, name))


TensorInfo = namedtuple('TensorInfo', ['name', 'shape', 'dtype'])


class Executor:

    def input_info(self):
        raise NotImplementedError()

    def output_info(self):
        raise NotImplementedError()

    def tensor_info(self):
        raise NotImplementedError()

    def __call__(self, inputs, output_names=None, collect_statistics=False):
        raise NotImplementedError()


class TFExecutor(Executor):

    def __init__(self, model_path):
        try:
            import tensorflow.compat.v1 as tf
        except ImportError:
            import tensorflow as tf
        from .io.tf.graphdef.protobuf import GraphDef

        self.Session = tf.Session

        graph_def = GraphDef()
        with open(model_path, 'rb') as file:
            graph_def.ParseFromString(file.read())

        self.graph = tf.Graph()
        with self.graph.as_default():
            tf.import_graph_def(graph_def, name='')

        ops = self.graph.get_operations()
        consumed = {tensor for op in ops for tensor in op.inputs}

        self.inputs = [op.outputs[0] for op in ops if op.type == 'Placeholder']
        self.outputs = [tensor for op in ops if len(op.inputs) for tensor in op.outputs
                        if tensor not in consumed and tensor.name.endswith(':0')]

    def input_info(self):
        return [TensorInfo(tensor.name, tuple(tensor.shape.as_list()), tensor.dtype.as_numpy_dtype)
                for tensor in self.inputs]

    def output_info(self):
        return [TensorInfo(tensor.name, tuple(tensor.shape.as_list()), tensor.dtype.as_numpy_dtype)
                for tensor in self.outputs]

    def tensor_info(self):
        tensors = [tensor for op in self.graph.get_operations() for tensor in op.outputs]
        return [TensorInfo(tensor.name, tuple(tensor.shape.as_list()), tensor.dtype.as_numpy_dtype)
                for tensor in tensors]

    def __call__(self, inputs, output_names=None, collect_statistics=False):
        ops = self.graph.get_operations()

        if output_names is not None:
            tensor_names = {tensor.name for op in ops for tensor in op.outputs}
            invalid = {name for name in output_names if name not in tensor_names}
            if len(invalid):
                raise ValueError('Invalid tensor name(s): {}'.format(invalid))

            outputs = {tensor.name: tensor for op in ops for tensor in op.outputs if tensor.name in output_names}
        else:
            outputs = {tensor.name: tensor for tensor in self.outputs}

        if collect_statistics:
            tensors = {tensor.name: tensor for op in ops for tensor in op.outputs if tensor.name.endswith(':0')}
            with self.Session(graph=self.graph) as sess:
                values = sess.run(tensors, feed_dict=inputs)

                outputs = {name: values[name] for name in outputs}

                stats = {}
                for name, array in six.iteritems(values):
                    stats[name] = compute_statistics(array)

                return outputs, stats
        else:
            with self.Session(graph=self.graph) as sess:
                outputs = sess.run(outputs, feed_dict=inputs)

            return outputs, None


class TFLiteExecutor(Executor):

    def __init__(self, model_path):
        try:
            import tensorflow.compat.v1 as tf
        except ImportError:
            import tensorflow as tf

        self.interpreter = tf.lite.Interpreter(model_path=model_path)
        self.interpreter.allocate_tensors()

    def input_info(self):
        return [TensorInfo(tensor['name'], tensor['shape'], tensor['dtype'])
                for tensor in self.interpreter.get_input_details()]

    def output_info(self):
        return [TensorInfo(tensor['name'], tensor['shape'], tensor['dtype'])
                for tensor in self.interpreter.get_output_details()]

    def tensor_info(self):
        return [TensorInfo(tensor['name'], tensor['shape'], tensor['dtype'])
                for tensor in self.interpreter.get_tensor_details()]

    def __call__(self, inputs, output_names=None, collect_statistics=False):
        for tensor in self.interpreter.get_input_details():
            self.interpreter.set_tensor(tensor['index'], inputs[tensor['name']])

        self.interpreter.invoke()

        if output_names is not None:
            tensor_names = {tensor['name'] for tensor in self.interpreter.get_tensor_details()}
            invalid = {name for name in output_names if name not in tensor_names}
            if len(invalid):
                raise ValueError('Invalid tensor name(s): {}'.format(invalid))

            outputs = {tensor['name']: self.interpreter.get_tensor(tensor['index'])
                       for tensor in self.interpreter.get_tensor_details()
                       if tensor['name'] in output_names}
        else:
            outputs = {tensor['name']: self.interpreter.get_tensor(tensor['index'])
                       for tensor in self.interpreter.get_output_details()}

        stats = {tensor['name']: compute_statistics(self.interpreter.get_tensor(tensor['index']))
                 for tensor in self.interpreter.get_tensor_details()} if collect_statistics else None

        return outputs, stats


class ONNXExecutor(Executor):

    def __init__(self, model_path, require_intermediates=False):
        import onnxruntime

        options = onnxruntime.SessionOptions()
        options.inter_op_num_threads = 1
        options.intra_op_num_threads = 1
        options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL

        self.session = onnxruntime.InferenceSession(model_path, sess_options=options,
                                                    providers=['CPUExecutionProvider'])
        self.inputs = [TensorInfo(tensor.name, tensor.shape, _onnx_dtype_to_numpy[tensor.type])
                       for tensor in self.session.get_inputs()]
        self.outputs = [TensorInfo(tensor.name, tensor.shape, _onnx_dtype_to_numpy[tensor.type])
                        for tensor in self.session.get_outputs()]

        if require_intermediates:
            import onnx
            from onnx.shape_inference import infer_shapes

            model = onnx.load_model(model_path)
            model = infer_shapes(model)

            for info in model.graph.value_info:
                output_info = model.graph.output.add()
                output_info.ParseFromString(info.SerializeToString())

            self.session = onnxruntime.InferenceSession(model.SerializeToString(), sess_options=options,
                                                        providers=['CPUExecutionProvider'])

    def input_info(self):
        return self.inputs

    def output_info(self):
        return self.outputs

    def tensor_info(self):
        return None

    def __call__(self, inputs, output_names=None, collect_statistics=False):
        if output_names is not None:
            inputs_as_outputs = {name: inputs[name] for name in output_names if name in inputs}
            output_names = [name for name in output_names if name not in inputs]
        else:
            output_names = [output.name for output in self.outputs]
            inputs_as_outputs = {}

        if collect_statistics:
            original_outputs = [output.name for output in self.outputs]
            fetch_names = [tensor.name for tensor in self.session.get_outputs()
                           if tensor.name not in original_outputs] + original_outputs
            values = self.session.run(fetch_names, inputs)
            outputs = {name: value for name, value in zip(fetch_names, values) if name in set(output_names)}

            stats = {}
            for name, value in zip(fetch_names, values):
                stats[name] = compute_statistics(value)
        else:
            values = self.session.run(output_names, inputs)
            outputs = {name: value for name, value in zip(output_names, values)}
            stats = None

        outputs.update(inputs_as_outputs)

        return outputs, stats


class NNEFExecutor(Executor):

    def __init__(self, model_path, custom_operators, decomposed):
        from .interpreter.pytorch import Interpreter
        self.interpreter = Interpreter(model_path, custom_operators=custom_operators, decomposed=decomposed)

    def input_info(self):
        return [TensorInfo(tensor.name, tensor.shape, _nnef_dtype_to_numpy[tensor.dtype])
                for tensor in self.interpreter.input_details()]

    def output_info(self):
        return [TensorInfo(tensor.name, tensor.shape, _nnef_dtype_to_numpy[tensor.dtype])
                for tensor in self.interpreter.output_details()]

    def tensor_info(self):
        return [TensorInfo(tensor.name, tensor.shape, _nnef_dtype_to_numpy[tensor.dtype])
                for tensor in self.interpreter.tensor_details()]

    def __call__(self, inputs, output_names=None, collect_statistics=False):
        inputs = [inputs[tensor.name] for tensor in self.interpreter.input_details()]
        if collect_statistics:
            return self.interpreter(inputs, output_names, collect_statistics)
        else:
            return self.interpreter(inputs, output_names, collect_statistics), None


def get_executor(format, model_path, require_intermediates, custom_operators, decomposed):
    if format == 'tf':
        return TFExecutor(model_path)
    elif format == 'tflite':
        return TFLiteExecutor(model_path)
    elif format == 'onnx':
        return ONNXExecutor(model_path, require_intermediates)
    elif format == 'nnef':
        return NNEFExecutor(model_path, custom_operators, decomposed)
    else:
        return None


def write_nnef_tensor(filename, value):
    with open(filename, 'wb') as file:
        dtype = _numpy_dtype_remap.get(value.dtype.type)
        if dtype is not None:
            value = value.astype(dtype)

        nnef.write_tensor(file, value)


def write_statistics(filename, statistics):
    statistics = {name: {'min': stats.min, 'max': stats.max, 'mean': stats.mean(), 'std': stats.std()}
                  for name, stats in six.iteritems(statistics)}

    with open(filename, 'w') as file:
        json.dump(statistics, file, indent=4)


def get_custom_operators(module_names):
    CUSTOM_OPERATORS = "CUSTOM_OPERATORS"

    operators = {}
    for module_name in module_names:
        module = importlib.import_module(module_name)
        if hasattr(module, CUSTOM_OPERATORS):
            operators.update(getattr(module, CUSTOM_OPERATORS))

    return operators


def batched_info(tensor_info, batch_size):
    for info in tensor_info:
        if info.shape[0] != batch_size and info.shape[0] != 1 and not isinstance(info.shape[0], str):
            raise ValueError('invalid input shape {} for batch size {}'.format(info.shape, batch_size))

    return [TensorInfo(name=info.name, shape=(batch_size, *info.shape[1:]), dtype=info.dtype)
            for info in tensor_info]


def accumulate_statistics(global_stats, local_stats):
    if global_stats is None:
        return local_stats
    for name, stats in six.iteritems(local_stats):
        global_stats[name] += stats
    return global_stats


def main(args):
    if args.input_path is not None:
        source = FileInputSource(args.input_path, args.io_transpose)
    elif args.random is not None:
        if args.batch_size == 0:
            print('batch-size must not be 0 when inputs are random generated', file=sys.stderr)
            return -1

        try:
            distribution = eval(args.random)
            if not _is_lambda(distribution):
                distribution = distribution()
            source = RandomInputSource(distribution)
        except Exception as e:
            print("Could not evaluate distribution: " + str(e), file=sys.stderr)
            return -1
    else:
        if not stdio.is_stdin_piped():
            print('Input must be piped', file=sys.stderr)
            return -1

        stdio.set_stdin_to_binary()
        source = StreamInputSource(sys.stdin, args.io_transpose)

    output_names = eval(args.output_names) if args.output_names is not None and args.output_names != "*" else args.output_names
    custom_operators = get_custom_operators(args.custom_operators) if args.custom_operators is not None else None

    if args.random is not None and args.seed is not None:
        np.random.seed(args.seed)

    collect_statistics = args.statistics is not None

    try:
        executor = get_executor(args.format, args.model, collect_statistics, custom_operators, args.decompose)

        if isinstance(output_names, dict):
            fetch_names = output_names.keys()
        elif output_names == "*":
            tensors = executor.tensor_info()
            fetch_names = [info.name for info in tensors] if tensors is not None else None
        else:
            fetch_names = output_names

        input_info = executor.input_info()
        if args.batch_size is not None:
            input_info = batched_info(input_info, args.batch_size)

        output_info = executor.output_info()

        inputs = {info.name: source(info.name, info.shape, info.dtype) for info in input_info}

        batch_size = args.batch_size
        if batch_size == 0:
            batch_size = next(iter(six.itervalues(inputs))).shape[0]
            if not all(input.shape[0] == batch_size for input in six.itervalues(inputs)):
                print('All inputs must have the same batch-size', file=sys.stderr)
                return -1

        if batch_size is not None and batch_size != 1:
            slices = {name: [] for name in fetch_names} if fetch_names is not None else \
                     {info.name: [] for info in output_info}
            stats = None

            for k in range(batch_size):
                slice_inputs = {name: np.expand_dims(data[k], axis=0) for name, data in six.iteritems(inputs)}
                slice_outputs, slice_stats = executor(slice_inputs, fetch_names, collect_statistics)

                for name, data in six.iteritems(slice_outputs):
                    slices[name].append(data)

                if collect_statistics:
                    stats = accumulate_statistics(stats, slice_stats)

            outputs = {name: np.concatenate(items, axis=0) for name, items in six.iteritems(slices)}
        else:
            outputs, stats = executor(inputs, fetch_names, collect_statistics)

    except ValueError as e:
        print(e, file=sys.stderr)
        return -1

    for name, value in six.iteritems(outputs):
        if needs_transpose(args.io_transpose, name):
            outputs[name] = transpose_channels_last_to_first(value)

    if isinstance(output_names, dict):
        outputs = {output_names[name]: value for name, value in six.iteritems(outputs)}

    if args.tensor_mapping is not None:
        with open(args.tensor_mapping) as file:
            tensor_mapping = json.load(file)

        if stats is not None:
            stats = {tensor_mapping.get(key, key): value for key, value in six.iteritems(stats)}

    if stats is not None:
        write_statistics(args.statistics, stats)
        print('Written {}'.format(args.statistics))

    if args.output_path is not None:
        for name, value in six.iteritems(outputs):
            filename = os.path.join(args.output_path, name + ".dat")
            write_nnef_tensor(filename, value)
            print('Written {}'.format(filename))
    else:
        if not stdio.is_stdout_piped():
            if collect_statistics:
                return 0

            print('Output must be piped', file=sys.stderr)
            return -1

        stdio.set_stdout_to_binary()

        for name, value in six.iteritems(outputs):
            nnef.write_tensor(sys.stdout, value)

    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('model', type=str,
                        help='The model to execute')
    parser.add_argument('--format', type=str, required=True, choices=['tf', 'tflite', 'onnx', 'nnef'],
                        help='The format of the model')
    parser.add_argument('--random', type=str, default=None,
                        help='Random distribution for input generation')
    parser.add_argument('--seed', type=int, default=None,
                        help='Random seed for input generation')
    parser.add_argument('--input-path', type=str, default=None,
                        help='Folder to read inputs from')
    parser.add_argument('--output-path', type=str, default=None,
                        help='Folder to save outputs into')
    parser.add_argument('--output-names', type=str, default=None,
                        help='The set (dict) of tensor names (to file names) considered as outputs to be saved. '
                             'Use * to save all tensors')
    parser.add_argument('--io-transpose', type=str, nargs='*', default=None,
                        help='The inputs/outputs to transpose from channels last to channels first dimension order')
    parser.add_argument('--decompose', type=str, nargs='*', default=None,
                        help='Names of operators to be decomposed by NNEF parser')
    parser.add_argument('--statistics', type=str, nargs='?', default=None, const='stats.json',
                        help='Calculate activations statistics and save to output path in json format')
    parser.add_argument('--custom-operators', type=str, nargs='+', default=None,
                        help='Module(s) containing custom operator code')
    parser.add_argument('--batch-size', type=int, default=None,
                        help='Specify batch-size for single-batch models')
    parser.add_argument('--tensor-mapping', type=str, default=None,
                        help='Use mapping of tensor names for statistics')
    exit(main(parser.parse_args()))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tflite

class TensorType(object):
    FLOAT32 = 0
    FLOAT16 = 1
    INT32 = 2
    UINT8 = 3
    INT64 = 4
    STRING = 5
    BOOL = 6
    INT16 = 7
    COMPLEX64 = 8
    INT8 = 9
    FLOAT64 = 10




# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .utils import stdio
import numpy as np
import argparse
import nnef
import sys
import skimage
import skimage.io
import skimage.color
import skimage.transform
import glob
import os


def transform_image(img, color, range, mean, std, size, dtype, data_format):
    img = img.astype(np.float32) / 255.0

    if color.upper() == 'RGB':
        img = img[..., (0, 1, 2)]  # remove alpha channel if present
    else:
        img = img[..., (2, 1, 0)]

    if range is not None:
        min = np.array(range[0], dtype=np.float32)
        max = np.array(range[1], dtype=np.float32)
        img *= max - min
        img += min

    if mean is not None:
        mean = np.array(mean, dtype=np.float32)
        img -= mean

    if std is not None:
        std = np.array(std, dtype=np.float32)
        img /= std

    if size is not None:
        img = skimage.transform.resize(img, size,
                                       preserve_range=True,
                                       anti_aliasing=True,
                                       mode='reflect')
    if dtype is not None:
        img = img.astype(dtype)

    if data_format.upper() == 'NCHW':
        img = img.transpose((2, 0, 1))

    return img


def main(args):
    if args.output is None:
        if not stdio.is_stdout_piped():
            print("Output must be piped", file=sys.stderr)
            return -1
        stdio.set_stdout_to_binary()

    images = []
    for pattern in args.images:
        filenames = sorted(glob.glob(os.path.expanduser(pattern)))
        assert filenames, "No files found for path: {}".format(pattern)
        for filename in filenames:
            img = skimage.img_as_ubyte(skimage.io.imread(filename))
            if len(img.shape) == 2:
                img = skimage.color.gray2rgb(img)

            img = transform_image(img, args.color, args.range, args.mean, args.std, args.size,
                                  np.dtype(args.dtype), args.format)
            images.append(img)

    if not all(img.shape == images[0].shape for img in images):
        print("The size of all images must be the same, or --size must be specified", file=sys.stderr)
        return -1

    tensor = np.stack(images)

    if args.output is not None:
        with open(args.output, 'wb') as file:
            nnef.write_tensor(file, tensor)
    else:
        nnef.write_tensor(sys.stdout, tensor)
    
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('images', type=str, nargs='+',
                        help='The path(s) of images to turn into a tensor; may include wildcard expressions')
    parser.add_argument('--size', type=int, nargs=2, default=None,
                        help='The spatial size of the resulting tensor')
    parser.add_argument('--dtype', type=str, default='float32',
                        help='The data-type of the resulting tensor')
    parser.add_argument("--color", type=str.upper, choices=['RGB', 'BGR'], default='RGB',
                        help="The resulting color-format")
    parser.add_argument("--format", type=str.upper, choices=['NCHW', 'NHWC'], default='NCHW',
                        help="The resulting data-format")
    parser.add_argument("--range", type=float, nargs=2, default=[0, 1],
                        help="Resulting range for representing the image")
    parser.add_argument("--mean", type=float, nargs='+', default=None,
                        help="Mean to subtract from the image; may be per-channel")
    parser.add_argument("--std", type=float, nargs='+', default=None,
                        help="Standard deviation to divide the image with; may be per-channel")
    parser.add_argument('--output', type=str, default=None,
                        help='File name to save the result into')
    exit(main(parser.parse_args()))



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..model.utils import replace_chain
from ..model.graph import *
from ..utils.types import from_numpy


class Optimizer:

    def __init__(self, custom_optimizers=None):
        self._custom_optimizers = custom_optimizers or {}

    def __call__(self, graph, only_required=False):
        replace_chain(graph, ['DEQUANTIZE'], self._eliminate_variable_dequantize)
        replace_chain(graph, ['SPACE_TO_BATCH_ND', {'CONV_2D', 'DEPTHWISE_CONV_2D'}, 'BATCH_TO_SPACE_ND'], self._replace_dilated_conv)
        replace_chain(graph, ['RESHAPE', 'RESHAPE', 'PACK', 'PACK', 'RESHAPE'], self._replace_resize_nearest)
        replace_chain(graph, ['SHAPE'], self._replace_const_shape)
        for chain, replacer in six.iteritems(self._custom_optimizers):
            replace_chain(graph, chain, replacer)
        return graph

    @staticmethod
    def _replace_resize_nearest(reshape1, reshape2, pack1, pack2, reshape3):
        def _all_inputs_same(op):
            return len(op.inputs) > 0 and all(tensor is op.inputs[0] for tensor in op.inputs)

        if not (reshape2.output.shape == reshape1.inputs[0].shape and
                _all_inputs_same(pack1) and _all_inputs_same(pack2) and
                len(pack1.inputs) == len(pack2.inputs)):
            return False

        input = reshape1.inputs[0]
        output = reshape3.output
        size = Tensor(input.graph, shape=(len(output.shape) - 2,), dtype=np.int32, data=np.array(output.shape[1:-1]),
                      name=output.name + '/size')
        Operation(input.graph,
                  type='RESIZE_NEAREST_NEIGHBOR',
                  inputs=(input, size),
                  outputs=output,
                  attribs={
                      'align_corners': False,
                      'half_pixel_centers': False,
                  })

    @staticmethod
    def _replace_dilated_conv(space_to_batch, conv, batch_to_space):
        if not Optimizer._is_constant(space_to_batch.inputs[1]) or not Optimizer._is_constant(batch_to_space.inputs[1]):
            return False

        block_shape1 = Optimizer._read_constant(space_to_batch.inputs[1])
        block_shape2 = Optimizer._read_constant(batch_to_space.inputs[1])

        if not np.all(block_shape1 == block_shape2):
            return False

        if conv.attribs['padding'] != 'VALID':
            return False

        strides = [1, conv.attribs['stride_h'], conv.attribs['stride_w'], 1]
        dilations = from_numpy(block_shape1)

        input = space_to_batch.inputs[0]
        filter = conv.inputs[1]
        output = batch_to_space.outputs[0]

        same_padding = Optimizer._is_same_padded(input.shape, output.shape, strides)

        if not same_padding:
            return False

        op = conv.copy_with(inputs=(input, filter, *conv.inputs[2:]), outputs=output, attribs=dict(conv.attribs))

        op.attribs['dilation_h_factor'] = dilations[0]
        op.attribs['dilation_w_factor'] = dilations[1]
        op.attribs['padding'] = 'SAME'

    @staticmethod
    def _is_constant(tensor):
        return tensor.producer is None and tensor.data is not None

    @staticmethod
    def _read_constant(tensor):
        return tensor.data

    @staticmethod
    def _is_same_padded(input, output, stride, is_nxc=True):
        rank = len(input)
        return all(output[i] == (input[i] + stride[i] - 1) // stride[i]
                   for i in (range(1, rank - 1) if is_nxc else range(2, rank)))

    @staticmethod
    def _eliminate_variable_dequantize(dequantize):
        if not Optimizer._is_constant(dequantize.input):
            return False

        variable = dequantize.input
        if variable.data is None:
            return False

        zero_point = variable.quant['zero_point']
        scale = variable.quant['scale']

        variable.data = (variable.data - zero_point) * scale
        variable.data = variable.data.astype(np.float32)
        variable.dtype = np.float32
        variable.quant = None

    @staticmethod
    def _replace_const_shape(shape):
        if shape.input.shape is not None:
            shape.output.data = np.array(shape.input.shape)



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..model.utils import bypass_and_remove, replace_chain
from ..model.utils import generate_tensor_names_from_op_type, generate_missing_tensor_names_from_op_type
from ..model.graph import *


class Optimizer:

    def __init__(self, keep_tensor_names=True, custom_optimizers=None, dequantize=False):
        self._keep_tensor_names = keep_tensor_names
        self._custom_optimizers = custom_optimizers or {}
        self._dequantize = dequantize

    def __call__(self, graph, only_required=False):
        self._fix_inputs_as_output(graph)
        self._fix_inputs_without_producer(graph)

        if not only_required:
            changed = True
            while changed:
                changed = False

                changed |= self._remove_identity_ops(graph, 'copy', lambda op: True)
                changed |= self._remove_identity_ops(graph, 'transpose', lambda op:
                    self._is_sorted(op.attribs['axes']))
                changed |= self._remove_identity_ops(graph, 'reshape', lambda op:
                    op.output.shape == op.input.shape)
                changed |= self._remove_identity_ops(graph, 'squeeze', lambda op:
                    op.attribs['axes'] == [])
                changed |= self._remove_identity_ops(graph, 'unsqueeze', lambda op:
                    op.attribs['axes'] == [])
                changed |= self._remove_identity_ops(graph, 'mul', lambda op:
                    self._is_constant(op.inputs[0], 1.0) or self._is_constant(op.inputs[1], 1.0))
                changed |= self._remove_identity_ops(graph, 'add', lambda op:
                    self._is_constant(op.inputs[0], 0.0) or self._is_constant(op.inputs[1], 0.0))
                changed |= self._remove_identity_ops(graph, ('box', 'debox', 'avg_pool', 'max_pool'), lambda op:
                    self._is_uniform(op.attribs['size'], 1) and
                    self._is_uniform(op.attribs['stride'], 1) and
                    self._is_uniform(op.attribs['dilation'], 1) and
                    self._is_uniform(op.attribs['padding'], 0))
                changed |= self._remove_identity_ops(graph,
                    ('nearest_downsample', 'area_downsample', 'nearest_upsample', 'multilinear_upsample'), lambda op:
                    self._is_uniform(op.attribs['factor'], 1))

                changed |= self._remove_inverse_ops(graph, 'squeeze', 'unsqueeze', lambda op1, op2:
                    op1.attribs['axes'] == op2.attribs['axes'])
                changed |= self._remove_inverse_ops(graph, 'unsqueeze', 'squeeze', lambda op1, op2:
                    op1.attribs['axes'] == op2.attribs['axes'])
                changed |= self._remove_inverse_ops(graph, 'transpose', 'transpose', lambda op1, op2:
                    self._is_sorted(Optimizer._permute(op1.attribs['axes'], op2.attribs['axes'])))

                changed |= self._merge_op_into_variables_and_constants(graph, 'transpose', lambda data, attribs:
                    data.transpose(attribs['axes']))
                changed |= self._merge_op_into_variables_and_constants(graph, 'reshape', lambda data, attribs:
                    data.reshape(attribs['shape']))
                changed |= self._merge_op_into_variables_and_constants(graph, 'squeeze', lambda data, attribs:
                    data.squeeze(tuple(attribs['axes'])))
                changed |= self._merge_op_into_variables_and_constants(graph, 'unsqueeze', lambda data, attribs:
                    data.reshape(self._unsqueeze_shape(data.shape, attribs['axes'])))

                changed |= self._merge_reshape_sequence(graph)

                changed |= replace_chain(graph, ['pad', {'conv', 'deconv', 'max_pool', 'avg_pool'}],
                                         self._merge_pad_with_sliding)
                changed |= replace_chain(graph, [{'mul', 'div'}, {'conv', 'deconv', 'linear'}],
                                         self._merge_mul_linear, allow_forks=True)
                changed |= replace_chain(graph, [{'conv', 'deconv', 'linear'}, {'add', 'sub'}],
                                         self._merge_linear_add)
                changed |= replace_chain(graph, [{'conv', 'deconv', 'linear'}, {'mul', 'div'}],
                                         self._merge_linear_mul)
                changed |= replace_chain(graph, ['matmul', {'add', 'sub'}],
                                         self._merge_matmul_bias)
                changed |= replace_chain(graph, [{'conv', 'deconv'}, 'batch_normalization'],
                                         self._merge_conv_batch_norm)
                changed |= replace_chain(graph, ['batch_normalization'], self._merge_batch_norm)
                changed |= replace_chain(graph, ['transpose', 'squeeze'], self._merge_transpose_squeeze)
                changed |= replace_chain(graph, ['reshape'], self._substitute_squeeze)

                for chain, replacer in six.iteritems(self._custom_optimizers):
                    changed |= replace_chain(graph, chain, replacer)

                changed |= self._remove_unused_variables_and_constants(graph)

        if self._keep_tensor_names:
            generate_missing_tensor_names_from_op_type(graph)
        else:
            generate_tensor_names_from_op_type(graph)

        if self._dequantize:
            Optimizer._dequantize_variables(graph)
            Optimizer._remove_quantization_attribs(graph)

        return graph

    @staticmethod
    def _fix_inputs_without_producer(graph):
        idx = 0
        for tensor in graph.inputs:
            if tensor.producer is None:
                cnt = len(graph.operations)
                Operation(tensor.graph, type='external', outputs=tensor,
                          attribs={'shape': list(tensor.shape), 'dtype': tensor.dtype})
                graph.move_operation(cnt, idx)
                idx += 1
        return idx > 0

    @staticmethod
    def _fix_inputs_as_output(graph):
        graph.outputs = [Optimizer._insert_copy(tensor) if tensor in graph.inputs else tensor
                         for tensor in graph.outputs]

    @staticmethod
    def _insert_copy(tensor, copy=None):
        if copy is None:
            copy = Tensor(tensor.graph, name=tensor.name + '_copy', dtype=tensor.dtype, shape=tensor.shape,
                          data=tensor.data, quant=tensor.quant)
        Operation(tensor.graph, type='copy', inputs=tensor, outputs=copy)
        return copy

    @staticmethod
    def _match_op_type(type, types):
        return type in types if isinstance(types, tuple) else type == types

    def _remove_identity_ops(self, graph, type, cond):
        changed = False
        for op in graph.operations:
            if self._match_op_type(op.type, type) and cond(op) and op.input.quant == op.output.quant:
                changed |= self._bypass_and_remove(graph, op)

        return changed

    def _merge_op_into_variables_and_constants(self, graph, type, func):
        changed = False
        for op in graph.operations:
            if (op.type == 'variable' or op.type == 'constant') and len(op.output.consumers) > 0:
                if self._all_consumers_same(op.output, type):
                    data = op.output.data if op.output.data is not None else np.zeros(op.output.shape)
                    attribs = op.output.consumers[0].attribs
                    data = func(data, attribs)
                    if op.output.data is not None:
                        op.output.data = data
                    op.output.shape = data.shape
                    op.attribs['shape'] = list(data.shape)
                    for consumer in list(op.output.consumers):  # copy the list before removals!
                        changed |= self._bypass_and_remove(graph, consumer)

        return changed

    def _remove_inverse_ops(self, graph, type1, type2, cond):
        changed = False
        for op in graph.operations:
            if op.type == type1 and len(op.output.consumers) == 1:
                consumer = op.output.consumer
                if consumer.type == type2 and cond(op, consumer):
                    changed |= self._bypass_and_remove(graph, op)
                    changed |= self._bypass_and_remove(graph, consumer)

        return changed

    def _merge_reshape_sequence(self, graph):
        changed = False
        for op in graph.operations:
            if op.type == 'reshape' and len(op.output.consumers) == 1:
                consumer = op.output.consumer
                if consumer.type == 'reshape':
                    new_shape = consumer.attribs['shape']
                    if any(s == 0 for s in new_shape):
                        old_shape = op.attribs['shape']
                        consumer.attribs['shape'] = [old_shape[i] if s == 0 else s for i, s in enumerate(new_shape)]
                    changed |= self._bypass_and_remove(graph, op)

        return changed

    def _bypass_and_remove(self, graph, op):
        if op.output in graph.outputs and (op.input in graph.inputs or op.input in graph.outputs):
            self._insert_copy(op.input, op.output)
            graph.remove_operation(op, unlink=True)
            return False
        else:
            bypass_and_remove(graph, op, remove_input_not_output=op.output in graph.outputs)
            return True

    @staticmethod
    def _is_channelwise_shape(shape):
        return len(shape) <= 1 or all(s == 1 or i == 1 for i, s in enumerate(shape))

    @staticmethod
    def _merge_linear_add(linear, add, type=None):
        bias = add.inputs[1] if add.inputs[0] == linear.output else add.inputs[0]
        if bias.data is None or not Optimizer._is_channelwise_shape(bias.shape):
            return False

        if len(linear.inputs) > 2 and linear.inputs[2].data is None:
            return None

        bias.data = Optimizer._squeeze_batch_and_spatial_dims(bias.data)
        bias.shape = bias.data.shape

        if add.type == 'sub':
            bias.data = -bias.data

        if len(linear.inputs) == 2:
            new_shape = (1, 1) if len(bias.shape) == 0 else (1, *bias.shape) if len(bias.shape) == 1 else None
            if new_shape is not None:
                bias.data = np.reshape(bias.data, newshape=new_shape)
                bias.shape = new_shape
        else:
            bias.data = linear.inputs[2].data + bias.data
            bias.shape = bias.data.shape

        Optimizer._ensure_variable_producer(bias, label=linear.output.name + '_bias')

        linear.copy_with(type=type or linear.type,
                         attribs=linear.attribs if type != 'linear' else {},
                         inputs=(linear.inputs[0], linear.inputs[1], bias),
                         outputs=add.output)

    @staticmethod
    def _merge_matmul_bias(matmul, add):
        bias = add.inputs[1] if add.inputs[0] == matmul.output else add.inputs[0]
        if not Optimizer._is_channelwise_shape(bias.shape):
            return False

        transposeA = matmul.attribs.get('transposeA') or False
        transposeB = matmul.attribs.get('transposeB') or False

        if transposeA:
            return False

        if not transposeB:
            producer = matmul.inputs[1].producer
            data = matmul.inputs[1].data
            if data is None or producer.type != 'variable':
                return False

            rank = len(data.shape)
            data = np.transpose(data, axes=list(range(rank - 2)) + [rank - 1, rank - 2])
            matmul.inputs[1].data = data
            producer.attribs['shape'] = list(data.shape)
            matmul.attribs['transposeB'] = True

        return Optimizer._merge_linear_add(matmul, add, type='linear')

    @staticmethod
    def _is_sorted(array):
        return all(array[i] <= array[i + 1] for i in range(len(array) - 1))

    @staticmethod
    def _all_consumers_same(tensor, type):
        attribs = tensor.consumers[0].attribs
        return all(consumer.type == type and consumer.attribs == attribs for consumer in tensor.consumers)

    @staticmethod
    def _unsqueeze_shape(shape, axes):
        for axis in axes:
            shape = shape[:axis] + (1,) + shape[axis:]
        return shape

    @staticmethod
    def _permute(items, perm):
        permuted = list(items)
        for i in range(len(perm)):
            permuted[i] = items[perm[i]]
        return type(items)(permuted)

    @staticmethod
    def _add_variable(graph, data, name, label=None):
        output = Tensor(graph, name=name, shape=data.shape, dtype=data.dtype.type, data=data)
        Operation(graph, type='variable', outputs=output, attribs={'shape': list(data.shape), 'label': label or name})
        return output

    @staticmethod
    def _ensure_variable_producer(tensor, label):
        if tensor.producer is None and len(tensor.shape) != 0:
            Operation(tensor.graph, type='variable', outputs=tensor,
                      attribs={'shape': list(tensor.shape), 'label': label})
        elif tensor.producer is not None:
            tensor.producer.attribs['shape'] = list(tensor.shape)

    @staticmethod
    def _merged_conv_batch_norm_params(weights, bias, mean, variance, offset, scale, epsilon, axis):
        std = np.sqrt(variance + epsilon)
        factor = scale / std
        new_weights = weights * np.reshape(factor, newshape=(1,) * axis + factor.shape + (1,) * (len(weights.shape) - axis - 1))
        new_bias = (bias - mean) * factor + offset
        return new_weights, new_bias

    @staticmethod
    def _merge_conv_batch_norm(conv, bn):
        if any(tensor.quant for tensor in conv.inputs) or any(tensor.quant for tensor in bn.inputs):
            return False

        if conv.inputs[1].data is None:
            return False

        weights, bias = Optimizer._merged_conv_batch_norm_params(conv.inputs[1].data,
                                    np.squeeze(conv.inputs[2].data if len(conv.inputs) > 2 else 0, axis=0),
                                    np.squeeze(bn.inputs[1].data, axis=0),
                                    np.squeeze(bn.inputs[2].data, axis=0),
                                    np.squeeze(bn.inputs[3].data if len(bn.inputs) > 3 else 0, axis=0),
                                    np.squeeze(bn.inputs[4].data if len(bn.inputs) > 4 else 1, axis=0),
                                    bn.attribs['epsilon'],
                                    axis=1 if conv.type == 'deconv' else 0)

        bias = np.expand_dims(bias, axis=0)

        conv.inputs[1].data = weights

        if len(conv.inputs) > 2:
            conv.inputs[2].data = bias
            conv.inputs[2].shape = bias.shape
            Optimizer._ensure_variable_producer(conv.inputs[2], label=conv.output.name + '_bias')
            conv.copy_with(outputs=bn.output)
        else:
            bias = Optimizer._add_variable(conv.graph, data=bias, name=conv.output.name + '_bias')
            conv.copy_with(inputs=(*conv.inputs[:2], bias), outputs=bn.output)

    @staticmethod
    def _merged_batch_norm_params(mean, variance, offset, scale, epsilon):
        std = np.sqrt(variance + epsilon)
        factor = scale / std
        return factor, offset - factor * mean

    @staticmethod
    def _merge_batch_norm(bn):
        if any(tensor.quant for tensor in bn.inputs):
            return False

        scale, offset = Optimizer._merged_batch_norm_params(
                                    bn.inputs[1].data,
                                    bn.inputs[2].data,
                                    bn.inputs[3].data if len(bn.inputs) > 3 else 0,
                                    bn.inputs[4].data if len(bn.inputs) > 4 else 1,
                                    bn.attribs['epsilon'])

        scale = Optimizer._add_variable(bn.graph, data=scale, name=bn.output.name + '_scale')
        offset = Optimizer._add_variable(bn.graph, data=offset, name=bn.output.name + '_offset')

        scaled = Tensor(graph=bn.graph, name=bn.output.name + '_scaled', shape=bn.output.shape, dtype=bn.output.dtype)

        Operation(graph=bn.graph, type='mul', inputs=(bn.inputs[0], scale), outputs=scaled)
        Operation(graph=bn.graph, type='add', inputs=(scaled, offset), outputs=bn.output)

    @staticmethod
    def _merge_mul_linear(mul, linear):
        which = 0 if mul.inputs[0].data is not None else 1
        other = 1 - which

        variable = mul.inputs[which]
        if variable.data is None or not Optimizer._is_channelwise_shape(variable.shape):
            return False

        if len(variable.shape) == 0:
            scale = np.expand_dims(variable.data, axis=0)
        elif len(variable.shape) >= 2:
            scale = Optimizer._squeeze_batch_and_spatial_dims(variable.data)

        weights = linear.inputs[1]
        if weights.data is None:
            return False

        rank = len(weights.shape)
        shape = scale.shape + (1,) * (rank - 1) if linear.type == 'deconv' else (1,) + scale.shape + (1,) * (rank - 2)
        scale = np.reshape(scale, newshape=shape)

        weights.data = weights.data * scale if mul.type != 'div' else weights.data / scale

        linear.copy_with(inputs=(mul.inputs[other], weights, *linear.inputs[2:]), outputs=linear.output)

    @staticmethod
    def _merge_linear_mul(linear, mul):
        variable = mul.inputs[1] if mul.inputs[0] == linear.output else mul.inputs[0]
        if variable.data is None or not Optimizer._is_channelwise_shape(variable.shape):
            return False

        if len(variable.shape) == 0:
            scale = np.expand_dims(variable.data, axis=0)
        elif len(variable.shape) >= 2:
            scale = Optimizer._squeeze_batch_and_spatial_dims(variable.data)

        negate = mul.type == 'div'

        weights = linear.inputs[1]
        if weights.data is None:
            return False

        if len(linear.inputs) > 2:
            bias = linear.inputs[2]
            if bias.data is None:
                return False

            bias.data = bias.data * scale if not negate else bias.data / scale
            bias.shape = bias.data.shape

            Optimizer._ensure_variable_producer(bias, label=linear.output.name + '_bias')

        rank = len(weights.shape)
        shape = (1,) + scale.shape + (1,) * (rank - 2) if linear.type == 'deconv' else scale.shape + (1,) * (rank - 1)
        scale = np.reshape(scale, newshape=shape)

        weights.data = weights.data * scale if not negate else weights.data / scale

        linear.copy_with(inputs=(linear.inputs[0], weights, *linear.inputs[2:]), outputs=mul.output)

    @staticmethod
    def _remove_unused_variables_and_constants(graph):
        ops = {op for op in graph.operations
               if (op.type == 'variable' or op.type == 'constant') and len(op.output.consumers) == 0}
        tensors = {op.output for op in ops}
        graph.remove_operations(ops, unlink=True)
        graph.remove_tensors(tensors)
        return len(ops) != 0

    @staticmethod
    def _merge_pad_with_sliding(pad, sliding):
        offset = 2 if sliding.type == 'conv' or sliding.type == 'deconv' else 0
        padding = pad.attribs['padding']

        if not all(p == 0 and q == 0 for p, q in sliding.attribs['padding']) or \
                len(padding) < offset or not all(p == 0 and q == 0 for p, q in padding[:offset]):
            return False

        attribs = dict(sliding.attribs)
        attribs['padding'] = pad.attribs['padding'][offset:]
        attribs['border'] = pad.attribs['border']

        sliding.copy_with(inputs=(pad.input, *sliding.inputs[1:]), attribs=attribs)

    @staticmethod
    def _squeeze_batch_and_spatial_dims(data):
        return np.squeeze(data, axis=(0,) + tuple(i for i in range(2, len(data.shape))))

    @staticmethod
    def _is_constant(tensor, value):
        if tensor.producer is not None and tensor.producer.name == 'constant':
            data = tensor.attribs['value']
        else:
            data = tensor.data

        return (not isinstance(tensor.data, np.ndarray) or data.shape == ()) and data == value

    @staticmethod
    def _is_uniform(array, value):
        return all(item == value for item in array)

    @staticmethod
    def _merge_transpose_squeeze(transpose, squeeze):
        transpose_axes = transpose.attribs['axes']
        squeeze_axes = squeeze.attribs['axes']

        squeezed = [x for i, x in enumerate(transpose_axes) if i not in squeeze_axes]
        is_identity = squeezed == list(range(len(squeezed)))

        if not is_identity:
            return False

        attribs = dict(squeeze.attribs)
        attribs['axes'] = [transpose_axes[x] for x in squeeze_axes]

        squeeze.copy_with(inputs=transpose.input, attribs=attribs)

    @staticmethod
    def _substitute_squeeze(reshape):
        input_shape = reshape.input.shape
        output_shape = reshape.output.shape

        if not len(output_shape) < len(input_shape):
            return False

        k = 0
        axes = []
        for i in range(len(input_shape)):
            if k < len(output_shape) and input_shape[i] == output_shape[k]:
                k += 1
            elif input_shape[i] == 1:
                axes.append(i)
            else:
                return False

        attribs = {'axes': axes}
        dtype = reshape.attribs.get('dtype')
        if dtype is not None:
            attribs['dtype'] = dtype

        Operation(reshape.graph, type='squeeze', name=reshape.name, inputs=reshape.input, outputs=reshape.output,
                  attribs=attribs)

    @staticmethod
    def _dequantize_variables(graph):
        for tensor in graph.tensors:
            if tensor.quant and tensor.data is not None:
                rank = len(tensor.data.shape)
                scale = Optimizer._ensure_quant_param_rank(tensor.quant.get('scale'), rank)
                zero_point = Optimizer._ensure_quant_param_rank(tensor.quant.get('zero_point'), rank)
                if isinstance(zero_point, np.ndarray):
                    assert Optimizer._broadcastable(zero_point.shape, tensor.shape), \
                        f"zero-point shape {zero_point.shape} cannot be broadcast to tensor shape {tensor.shape} " \
                        f"for tensor '{tensor.name}'"
                if isinstance(scale, np.ndarray):
                    assert Optimizer._broadcastable(scale.shape, tensor.shape), \
                        f"scale shape {scale.shape} cannot be broadcast to tensor shape {tensor.shape} " \
                        f"for tensor '{tensor.name}'"
                if scale is not None and not Optimizer._is_zero(scale):
                    dequantized = (tensor.data - zero_point) * scale
                    tensor.data = dequantized.astype(np.float32)
                    tensor.quant = None

    @staticmethod
    def _remove_quantization_attribs(graph):
        for tensor in graph.tensors:
            tensor.quant = None

    @staticmethod
    def _ensure_quant_param_rank(param, rank, offset=0):
        return np.reshape(param, newshape=(1,) * offset + param.shape + (1,) * (rank - 1 - offset)) \
            if isinstance(param, np.ndarray) and len(param.shape) == 1 else param

    @staticmethod
    def _broadcastable(x, y):
        return all(xi == yi or xi == 1 for xi, yi in zip(x, y))

    @staticmethod
    def _is_zero(value):
        return np.all(value == 0) if isinstance(value, np.ndarray) else value == 0



# Copyright (c) 2020 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function, absolute_import
from .converter import ConverterToNNEF as _Converter, Transform, ConversionError
from ..model.utils import generate_tensor_names_from_op_type
from ..model import Tensor
from ..utils import types
from collections import OrderedDict
import numpy as np
import copy
from nnef.shapes import pool_shape, reduce_shape


_LP_POOL_FRAGMENT = """
fragment lp_pool( 
    input: tensor<scalar>,
    size: integer[],
    border: string = 'constant',
    padding: (integer, integer)[] = [],
    stride: integer[] = [],
    dilation: integer[] = [],
    p: scalar = 2.0 ) 
-> ( output: tensor<scalar> )
{
    powered = pow(abs(input), p);
    summed = box(powered, size = size, border = border, padding = padding, stride = stride, dilation = dilation);
    output = pow(summed, 1.0 / p);
}
"""

_LP_REDUCE_FRAGMENT = """
fragment lp_reduce( 
    input: tensor<scalar>,
    axes: integer[],
    p: scalar = 2.0 ) 
-> ( output: tensor<scalar> )
{
    powered = pow(abs(input), p);
    summed = sum_reduce(powered, axes = axes);
    output = pow(summed, 1.0 / p);
}
"""

_MEAN_VARIANCE_NORMALIZATION_FRAGMENT = """
fragment mean_variance_normalization( 
    input: tensor<scalar>,
    scale: tensor<scalar>,
    offset: tensor<scalar>,
    axes: integer[],
    epsilon: scalar = 1e-5 ) 
-> ( output: tensor<scalar> )
{
    mean, variance = moments(input, axes = axes);
    output = scale * (input - mean) / sqrt(variance + epsilon) + offset;
}
"""

_LSTM_STEP_FRAGMENT = """
fragment lstm_step(
    x: tensor<scalar>,
    h: tensor<scalar>,
    c: tensor<scalar>,
    W: tensor<scalar>,
    R: tensor<scalar>,
    B: tensor<scalar> )
-> ( h_out: tensor<scalar>,
    c_out: tensor<scalar> )
{
    [Wb, Rb] = split(B, axis = 1, ratios = [1, 1]);
    z = linear(x, W, Wb) + linear(h, R, Rb);
    [i, f, g, o] = split(z, axis = 1, ratios=[1, 1, 1, 1]);
    c_out = sigmoid(f) * c + sigmoid(i) * tanh(g);
    h_out = sigmoid(o) * tanh(c_out);
}
"""

_LSTM_LOOP_FRAGMENT = """
fragment lstm_loop(
    X: tensor<scalar>,
    W: tensor<scalar>,
    R: tensor<scalar>,
    B: tensor<scalar>,
    h0: tensor<scalar>,
    c0: tensor<scalar>,
    steps: integer,
    index: integer = 0,
    axis: integer = 0 )
-> ( hn: tensor<scalar>, cn: tensor<scalar> )
{
    x0 = squeeze(slice(X, axes = [axis], begin = [index], end = [index + 1]), axes = [axis]);
    h1, c1 = lstm_step(x0, h0, c0, W, R, B);
    hn, cn = lstm_loop(X, W, R, B, h1, c1, index = index + 1, steps=steps) if index + 1 < steps else (h1, c1);
}
"""

_ERF_FRAGMENT = """
fragment erf( x: tensor<scalar> ) -> ( y: tensor<scalar> )
{
    t = 1.0 / (1.0 + 0.3275911 * abs(x));
    z = 1.0 - (((((1.061405429 * t + -1.453152027) * t) + 1.421413741) * t + -0.284496736) * t + 0.254829592) * t * exp(-x * x);
    y = sign(x) * z;
}
"""

_MISH_FRAGMENT = """
fragment mish( x: tensor<scalar> ) -> ( y: tensor<scalar> )
{
    y = x * tanh(ln(1 + exp(x)));
}
"""

_INT_MAX = 2 ** 31 - 1


class Converter(_Converter):

    @staticmethod
    def defined_operations():
        return {
            'lp_pool': _LP_POOL_FRAGMENT,
            'lp_reduce': _LP_REDUCE_FRAGMENT,
            'mean_variance_normalization': _MEAN_VARIANCE_NORMALIZATION_FRAGMENT,
            'lstm_step': _LSTM_STEP_FRAGMENT,
            'lstm_loop': _LSTM_LOOP_FRAGMENT,
            'erf': _ERF_FRAGMENT,
            'mish': _MISH_FRAGMENT,
        }

    @staticmethod
    def defined_operation_dependencies():
        return {
            'lstm_loop': ['lstm_step'],
        }

    @staticmethod
    def defined_shapes():
        return {
            'lp_pool': pool_shape,
            'lp_reduce': reduce_shape,
            'mean_variance_normalization': lambda input, scale, offset, **kwargs: input,
            'lstm_step': lambda x, h, c, W, R, B: (h, c),
            'lstm_loop': lambda X, W, R, B, h, c, **kwargs: (h, c),
            'erf': lambda x: x,
            'mish': lambda x: x,
        }

    def __init__(self, custom_transforms=None, custom_functions=None, mirror_unsupported=False, keep_io_names=False,
                 infer_shapes=False, custom_shapes=None, io_transpose=False):
        _Converter.__init__(self, transforms=self.merge_transforms(_Transforms, custom_transforms),
                            functions=custom_functions,
                            mirror_unsupported=mirror_unsupported,
                            infer_shapes=infer_shapes,
                            custom_shapes=dict(**self.defined_shapes(), **custom_shapes or {}))
        self._keep_io_names = keep_io_names
        self._io_transpose = io_transpose

    def __call__(self, graph):
        graph = _Converter.__call__(self, graph)
        self.remove_unused_constants(graph)
        self.inline_scalar_constants(graph)
        self.convert_constants_to_variables(graph)
        self._ensure_valid_ids(graph)
        if self._io_transpose is not False:
            self._transpose_inputs(graph)
            self._transpose_outputs(graph)
            graph.sort()
        generate_tensor_names_from_op_type(graph, keep_io_names=self._keep_io_names)
        return graph

    def _prepare(self, graph):
        self._insert_externals_and_constants(graph)

    def _is_constant(self, tensor):
        if tensor.producer:
            return tensor.producer.type == 'Constant'
        else:
            return tensor.data is not None

    def _read_constant(self, tensor, type=None):
        if tensor.producer and tensor.producer.type == 'Constant':
            value = tensor.producer.attribs['value']
        elif not tensor.producer:
            value = tensor.data
        else:
            raise ConversionError('trying to evaluate non-constant tensor')

        return types.from_numpy(value, type=type) if isinstance(value, np.ndarray) else types.cast(value, type=type)

    def _needs_io_transpose(self, tensor):
        if tensor.rank <= 2:
            return False
        if isinstance(self._io_transpose, bool):
            return self._io_transpose
        else:
            return tensor.name in self._io_transpose

    def _transpose_inputs(self, graph):
        inputs = [self._transpose_input(tensor) if self._needs_io_transpose(tensor) else tensor
                  for tensor in graph.inputs]

        if self._keep_io_names:
            for i in range(len(inputs)):
                if inputs[i] is not graph.inputs[i]:
                    inputs[i].name = graph.inputs[i].name

        graph.inputs = inputs

    def _transpose_outputs(self, graph):
        outputs = [self._transpose_output(tensor) if self._needs_io_transpose(tensor) else tensor
                   for tensor in graph.outputs]

        if self._keep_io_names:
            for i in range(len(outputs)):
                if outputs[i] is not graph.outputs[i]:
                    outputs[i].name = graph.outputs[i].name

        graph.outputs = outputs

    def _transpose_input(self, tensor):
        external = tensor.producer
        external.outputs = self._post_transpose(tensor, self.ncx_to_nxc_perm(tensor.rank))
        external.attribs['shape'] = list(self.nxc_to_ncx(tensor.shape))
        return external.output

    def _transpose_output(self, tensor):
        return self._pre_transpose(tensor, self.nxc_to_ncx_perm(tensor.rank))

    @staticmethod
    def _interleave(items):
        return [item[0] for item in items] + [item[1] for item in items]

    @staticmethod
    def _uninterleave(items):
        count = len(items) // 2
        return list(zip(items[:count], items[count:]))

    def convert_padding(self, pads, auto_pad, output_padding, rank):
        if auto_pad == "NOTSET" or auto_pad == "SAME_LOWER":
            padding = self._uninterleave(pads)
            if output_padding is not None:
                for i in range(len(padding)):
                    padding[i] = (padding[i][0], padding[i][1] - output_padding[i])
            return [(0, 0)] * (rank - len(padding)) + padding
        elif auto_pad == "VALID":
            padding = [(0, 0,)] * rank
            if output_padding is not None:
                offs = rank - len(output_padding)
                for i in range(len(output_padding)):
                    padding[i + offs] = (padding[i + offs][0], padding[i + offs][1] - output_padding[i])
            return padding
        elif auto_pad == "SAME_UPPER":
            return []
        else:
            assert False

    def convert_pads(self, pads):
        return self._uninterleave(pads)

    def squeeze_input(self, tensor, axes, keep_dims=False):
        return self._pre_squeeze(tensor, axes=axes) if not keep_dims and len(axes) else tensor

    def squeeze_output(self, tensor, axes, keep_dims=False):
        return self._post_squeeze(tensor, axes=axes) if not keep_dims and len(axes) else tensor

    def unsqueeze_input(self, tensor, axes, keep_dims=False):
        return self._pre_unsqueeze(tensor, axes=axes) if not keep_dims and len(axes) else tensor

    def unsqueeze_output(self, tensor, axes, keep_dims=False):
        return self._post_unsqueeze(tensor, axes=axes) if not keep_dims and len(axes) else tensor

    def unsqueeze_vector(self, tensor):
        original = self._tensor_map[tensor]
        if self._is_constant(original) and len(original.consumers) == 1:
            self._transform_constant(tensor, lambda data: np.expand_dims(data, 0))
            return tensor
        else:
            return self.unsqueeze_input(tensor, axes=[0])

    def bias_add(self, output, bias):
        if bias.rank == 0 and bias.data == 0:
            return output

        input = Tensor(output.graph, dtype=output.dtype, shape=output.shape, quant=copy.deepcopy(output.quant))
        self._bias_operation(input, output, bias)
        return input

    def lower_pads(self, input_size, filter_size, output_size, stride, dilation):
        rank = len(input_size)
        total = [None] * rank
        for i in range(rank):
            dilated_size = (filter_size[i] - 1) * dilation[i] + 1
            total[i] = max((input_size[i] // stride[i] - 1) * stride[i] + dilated_size - output_size[i], 0)
        pads = [(t // 2, t - t // 2) for t in total]
        return self._interleave(pads)

    def broadcast(self, tensor, rank):
        return self.unsqueeze_input(tensor, axes=list(range(rank - tensor.rank))) if tensor.rank > 0 else tensor

    def ensure_list(self, arg):
        return [arg] if not isinstance(arg, list) else arg

    def ensure_scalar(self, arg):
        return arg[0] if isinstance(arg, list) and len(arg) == 1 else arg

    def limit_range(self, x):
        return _INT_MAX if x > _INT_MAX else -_INT_MAX if x < -_INT_MAX else x

    def is_unused(self, tensor):
        if len(tensor.name) == 0:
            return True
        original = self._tensor_map[tensor]
        return len(original.consumers) == 0


_Transforms = Converter.unpack_transforms({
    ('Conv', 'ConvTranspose'):
        Transform(
            type=('conv', 'deconv'),
            defaults={
                'strides': '![1] * (I[0].rank - 2)',
                'dilations': '![1] * (I[0].rank - 2)',
                'pads': '![0, 0] * (I[0].rank - 2)',
                'auto_pad': "NOTSET",
                'group': 1,
                'output_shape': None,
                'output_padding': None,
            },
            using={
                '_pads': '!lower_pads(I[0].shape[2:], I[1].shape[2:], O[0].shape[2:], strides, dilations)'
                         ' if auto_pad == "SAME_LOWER" else pads',
            },
            inputs=(
                '!I[0]',
                '!I[1]',
                '!unsqueeze_vector(I[2]) if len(I) > 2 else None',
            ),
            outputs='!O[0]',
            attribs={
                'stride': '!strides',
                'dilation': '!dilations',
                'padding': '!convert_padding(_pads, auto_pad, output_padding, I[0].rank - 2)',
                'groups': '!group',
                'output_shape': '!output_shape',
            }
        ),
    ('MaxPool', 'AveragePool', 'LpPool'):
        Transform(
            type=('max_pool', 'avg_pool', 'lp_pool'),
            defaults={
                'strides': '![1] * (I[0].rank - 2)',
                'dilations': '![1] * (I[0].rank - 2)',
                'pads': '![0, 0] * (I[0].rank - 2)',
                'auto_pad': "NOTSET",
                'ceil_mode': 0,
                'storage_order': 0,
                'count_include_pad': 0,
            },
            cond={
                '!ceil_mode == 0': 'ceil_mode must be 0',
                '!storage_order == 0': 'storage_order must be 0',
            },
            using={
                '_pads': '!lower_pads(I[0].shape[2:], kernel_shape, O[0].shape[2:], strides, dilations)'
                         ' if auto_pad == "SAME_LOWER" else pads',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'size': '![1, 1] + kernel_shape',
                'stride': '![1, 1] + strides',
                'dilation': '![1, 1] + dilations',
                'padding': '!convert_padding(_pads, auto_pad, None, I[0].rank)',
                'border': '!"constant" if count_include_pad else "ignore"',
            }
        ),
    ('GlobalMaxPool', 'GlobalAveragePool', 'GlobalLpPool'):
        Transform(
            type=('max_reduce', 'mean_reduce', 'lp_reduce'),
            defaults={
                'p': 2,
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!list(range(2, I[0].rank))',
                'p': '!float(p) if _type_ == "GlobalLpPool" else None',
            }
        ),
    ('ReduceMin', 'ReduceMax', 'ReduceMean', 'ReduceSum', 'ReduceL1', 'ReduceL2'):
        Transform(
            type=('min_reduce', 'max_reduce', 'mean_reduce', 'sum_reduce', 'lp_reduce', 'lp_reduce'),
            defaults={
                'keepdims': 1,
            },
            inputs='!I[0]',
            outputs='!squeeze_output(O[0], axes, keepdims)',
            attribs={
                'axes': '!ensure_positive(axes, I[0].rank)',
                'p': '!1.0 if _type_ == "ReduceL1" else 2.0 if _type_ == "ReduceL2" else None',
            }
        ),
    ('ArgMin', 'ArgMax'):
        Transform(
            type=('argmin_reduce', 'argmax_reduce'),
            defaults={
                'axis': 0,
                'keepdims': 1,
                'select_last_index': 0,
            },
            using={
                'axes': '![ensure_positive(axis, I[0].rank)]',
            },
            cond={
                '!select_last_index == 0': 'select_last_index must be 0',
            },
            inputs='!I[0]',
            outputs='!squeeze_output(O[0], axes, keepdims)',
            attribs={
                'axes': '!axes',
            }
        ),
    'BatchNormalization':
        Transform(
            type='batch_normalization',
            defaults={
                'epsilon': 1e-5,
                'spatial': 1,
            },
            inputs=(
                '!I[0]',
                '!unsqueeze_vector(I[3])',
                '!unsqueeze_vector(I[4])',
                '!unsqueeze_vector(I[2])',
                '!unsqueeze_vector(I[1])',
            ),
            outputs='!O[0]',
            attribs={
                'epsilon': '!epsilon',
            }
        ),
    ('Relu', 'Sigmoid', 'Tanh', 'Softplus', 'Selu', 'Not', 'Identity', 'Elu', 'Erf', 'Mish', 'Abs', 'Sign',
     'Sin', 'Cos', 'Tan', 'Asin', 'Acos', 'Atan', 'Sinh', 'Cosh', 'Tanh', 'Asinh', 'Acosh', 'Atanh',
     'Exp', 'Log', 'Neg', 'Sqrt', 'Ceil', 'Floor', 'Round'):
        Transform(
            type=('relu', 'sigmoid', 'tanh', 'softplus', 'selu', 'not', 'copy', 'elu', 'erf', 'mish', 'abs', 'sign',
                  'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
                  'exp', 'log', 'neg', 'sqrt', 'ceil', 'floor', 'round'),
            inputs='!I[0]',
            outputs='!O[0]',
        ),
    ('Add', 'Sub', 'Mul', 'Div', 'Pow', 'Min', 'Max', 'And', 'Or',
     'Equal', 'Less', 'Greater', 'LessOrEqual', 'GreaterOrEqual'):
        Transform(
            type=('add', 'sub', 'mul', 'div', 'pow', 'min', 'max', 'and', 'or', 'eq', 'lt', 'gt', 'le', 'ge'),
            inputs=(
                '!broadcast(I[0], O[0].rank)',
                '!broadcast(I[1], O[0].rank)',
            ),
            outputs='!O[0]',
        ),
    'LeakyRelu':
        Transform(
            type='leaky_relu',
            defaults={
                'alpha': 0.01,
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'alpha': '!alpha',
            }
        ),
    'PRelu':
        Transform(
            type='prelu',
            inputs=(
                '!I[0]',
                '!broadcast(I[1], I[0].rank)',
            ),
            outputs='!O[0]',
        ),
    'Transpose':
        Transform(
            type='transpose',
            defaults={
                'perm': '!list(reversed(range(I[0].rank)))',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!ensure_positive(perm, I[0].rank)',
            }
        ),
    'Reshape':
        Transform(
            type='reshape',
            defaults={
                'shape': '!as_const(I[1])',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'shape': '!flexible_batch(shape, I[0].shape[0])',
            }
        ),
    'Flatten':
        Transform(
            type='reshape',
            defaults={
                'axis': 1,
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'shape': '![0] * axis + [-1]',
            }
        ),
    'Squeeze':
        Transform(
            type='squeeze',
            defaults={
                'axes': '!as_const(I[1])',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!ensure_positive(axes, I[0].rank)',
            }
        ),
    'Unsqueeze':
        Transform(
            type='unsqueeze',
            defaults={
                'axes': '!as_const(I[1])',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!ensure_positive(axes, O[0].rank)',
            }
        ),
    'MatMul':
        Transform(
            type='matmul',
            inputs=(
                '!broadcast(I[0], O[0].rank)',
                '!broadcast(I[1], O[0].rank)',
            ),
            outputs=(
                '!O[0]',
            ),
        ),
    'Gemm':
        Transform(
            type='!"linear" if is_linear else "matmul"',
            defaults={
                'alpha': 1.0,
                'beta': 1.0,
                'transA': 0,
                'transB': 0,
            },
            cond={
                '!alpha == 1.0': 'alpha must be 1',
                '!beta == 1.0 or len(I) == 2': 'beta must be 1',
            },
            using={
                'is_linear': '!len(I) > 2 and I[2].rank == 1 and transB',
                'bias': '!broadcast(I[2], O[0].rank) if len(I) > 2 and not is_linear else None',
            },
            inputs=(
                '!I[0]',
                '!I[1]',
                '!unsqueeze_vector(I[2]) if is_linear else None',
            ),
            outputs='!O[0] if is_linear or bias is None else bias_add(O[0], bias)',
            attribs={
                'transposeA': '!bool(transA) if not is_linear else None',
                'transposeB': '!bool(transB) if not is_linear else None',
            }
        ),
    'LRN':
        Transform(
            type='local_response_normalization',
            defaults={
                'alpha': 0.0001,
                'beta': 0.75,
                'bias': 1.0,
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'alpha': '!alpha',
                'beta': '!beta',
                'bias': '!bias',
                'size': '![1, size] + [1] * (I[0].rank - 2)',
            }
        ),
    'Concat':
        Transform(
            type='concat',
            defaults={
                'axis': 1,
            },
            inputs=['!I[:]'],
            outputs='!O[0]',
            attribs={
                'axis': '!ensure_positive(axis, O[0].rank)',
            }
        ),
    'Split':
        Transform(
            type='split',
            defaults={
                'axis': 0,
                'split': '!as_const(I[1])',
            },
            inputs='!I[0]',
            outputs=['!O[:]'],
            attribs={
                'axis': '!ensure_positive(axis, I[0].rank)',
                'ratios': '!split',
            }
        ),
    'Dropout':
        Transform(
            type='copy',
            inputs='!I[0]',
            outputs='!O[0]',
        ),
    'Softmax':
        Transform(
            type='softmax',
            defaults={
                'axis': 1,
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '![ensure_positive(axis, I[0].rank)]',
            }
        ),
    'Sum':
        Transform(
            type='add_n',
            inputs=['!I[:]'],
            outputs='!O[0]',
        ),
    'Where':
        Transform(
            type='select',
            inputs=(
                '!broadcast(I[0], O[0].rank)',
                '!broadcast(I[1], O[0].rank)',
                '!broadcast(I[2], O[0].rank)',
            ),
            outputs='!O[0]',
        ),
    'Clip':
        Transform(
            type='!"max" if I[2].name == "" else "min" if I[1].name == "" else "clamp"',
            inputs=(
                '!I[0]',
                '!I[1] if I[1].name != "" else None',
                '!I[2] if I[2].name != "" else None',
            ),
            outputs='!O[0]',
        ),
    'Pad':
        Transform(
            type='pad',
            defaults={
                'mode': "constant",
                'value': 0.0,
            },
            using={
                'constant_value': '!ensure_scalar(as_const(I[2])) if len(I) > 2 else value',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'padding': '!convert_pads(as_const(I[1]) if len(I) > 1 else pads)',
                'value': '!constant_value',
                'border': '!"replicate" if mode == "edge" else mode',
            }
        ),
    'Tile':
        Transform(
            type='tile',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'repeats': '!as_const(I[1])',
            }
        ),
    'Expand':
        Transform(
            type='tile',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'repeats': '![O[0].shape[i] // I[0].shape[i] for i in range(I[0].rank)]',
            }
        ),
    'Slice':
        Transform(
            type='slice',
            using={
                'axes': '!as_const(I[3]) if len(I) > 3 else list(range(I[0].rank))',
                'starts': '![limit_range(x) for x in as_const(I[1])]',
                'ends': '![limit_range(x) for x in as_const(I[2])]',
                'steps': '!as_const(I[4]) if len(I) > 4 else None',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '!ensure_positive(axes, I[0].rank)',
                'begin': '!starts',
                'end': '!ends',
                'stride': '!steps',
            }
        ),
    'LpNormalization':
        Transform(
            type='!"l1_normalization" if p == 1 else "l2_normalization"',
            defaults={
                'axis': -1,
                'p': 2,
            },
            cond={
                '!p == 1 or p == 2': 'p must be 1 or 2',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'axes': '![ensure_positive(axis, I[0].rank)]',
            }
        ),
    'MeanVarianceNormalization':
        Transform(
            type='mean_variance_normalization',
            defaults={
                'axes': [0, 2, 3],
            },
            inputs=(
                '!I[0]',
                '!as_tensor(1.0, np.float32, inline=True)',
                '!as_tensor(0.0, np.float32, inline=True)',
            ),
            outputs='!O[0]',
            attribs={
                'axes': '!axes',
                'epsilon': 0.0,
            }
        ),
    'InstanceNormalization':
        Transform(
            type='mean_variance_normalization',
            defaults={
                'epsilon': 1e-5,
            },
            inputs=(
                '!I[0]',
                '!unsqueeze_vector(I[1])',
                '!unsqueeze_vector(I[2])',
            ),
            outputs='!O[0]',
            attribs={
                'axes': '!list(range(2, I[0].rank))',
                'epsilon': '!epsilon',
            }
        ),
    'Upsample':
        Transform(
            type='!"nearest_upsample" if mode == "nearest" else "multilinear_upsample"',
            defaults={
                'mode': "nearest",
                'scales': '!as_const(I[1])',
            },
            cond={
                '!scales[0] == 1 and scales[1] == 1': 'scales must be 1 in batch and channel dimensions',
                '!all(int(s) == s for s in scales[2:])': 'scales must be integers in all dimensions',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'factor': '![int(s) for s in scales[2:]]',
                'method': '!"asymmetric" if mode == "linear" else None',
            }
        ),
    'Resize':
        Transform(
            type='!("nearest_downsample" if downsample else "nearest_upsample") if mode == "nearest" else'
                 ' "multilinear_upsample"',
            defaults={
                'mode': "nearest",
                'coordinate_transformation_mode': "half_pixel",
            },
            using=OrderedDict([
                ('scales', '!as_const(I[1 if len(I) == 2 else 2])'),
                ('sizes', '!as_const(I[3]) if len(I) > 3 else'
                          ' [int(I[0].shape[i] * scales[i]) for i in range(I[0].rank)]'),
                ('upsample', '!is_integer_upsample(I[0].shape, sizes)'),
                ('downsample', '!is_integer_downsample(I[0].shape, sizes)'),
            ]),
            cond={
                '!mode == "nearest" or mode == "linear"':
                    'mode must be one of "nearest", "linear"',
                '!upsample or downsample if mode == "nearest" else True':
                    "nearest resize must be integer up-sample or down-sample",
                '!upsample if mode == "linear" else True':
                    'linear resize must be integer up-sample',
                '!sizes[0] == I[0].shape[0] and sizes[1] == I[0].shape[1]':
                    'batch and channel dimensions must be preserved',
                '!coordinate_transformation_mode == "half_pixel" or'
                ' coordinate_transformation_mode == "pytorch_half_pixel" or'
                ' coordinate_transformation_mode == "asymmetric" or'
                ' coordinate_transformation_mode == "align_corners"':
                    'coordinate_transformation_mode must be one of'
                    ' "half_pixel", "pytorch_half_pixel", "asymmetric", "align_corners"',
            },
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'factor': '!upsample_factor(I[0].shape[2:], sizes[2:]) if upsample else'
                          ' downsample_factor(I[0].shape[2:], sizes[2:])',
                'method': '!("aligned" if coordinate_transformation_mode == "align_corners" else'
                          ' "symmetric" if coordinate_transformation_mode == "half_pixel" or '
                          ' coordinate_transformation_mode == "pytorch_half_pixel" else'
                          ' "asymmetric") if mode == "linear" else None',
            }
        ),
    'Constant':
        Transform(
            type='constant',
            outputs='!O[0]',
            attribs={
                'value': '!ensure_list(from_numpy(value)) if len(value.shape) <= 1 and int(np.prod(value.shape)) <= 10 '
                         'else value',
                'shape': '!list(value.shape)',
                'dtype': '!value.dtype',
            }
        ),
    'Gather':
        Transform(
            type='gather',
            inputs=('!I[0]', '!I[1]'),
            outputs='!O[0]',
            defaults={
                'axis': 0,
            },
            attribs={
                'axis': '!ensure_positive(axis, I[0].rank)',
            },
        ),
    'Cast':
        Transform(
            using={
                'same_type': '!nnef_dtype(O[0].dtype) == nnef_dtype(I[0].dtype)',
            },
            type='!"copy" if same_type else "cast"',
            inputs='!I[0]',
            outputs='!O[0]',
            attribs={
                'dtype': '!O[0].dtype if not same_type else None',
            },
        ),
    'LSTM':
        Transform(
            cond={
                '!direction == "forward"': 'direction must be "forward"',
                '!is_unused(O[0])': 'first output must not have consumer operations',
                '!len(I[4].name) == 0': 'sequence_lens must not be defined',
            },
            defaults={
                'layout': 0,
            },
            using={
                'seq_axis': '!0 if layout == 0 else 1',
                'dir_axis': '!0 if layout == 0 else 2',
            },
            type='lstm_loop',
            inputs=(
                '!I[0]',                                    # X
                '!squeeze_input(I[1], axes=[0])',           # W
                '!squeeze_input(I[2], axes=[0])',           # R
                '!I[3]',                                    # B
                '!squeeze_input(I[5], axes=[dir_axis])',    # h_0
                '!squeeze_input(I[6], axes=[dir_axis])',    # c_0
            ),
            outputs=(
                '!unsqueeze_output(O[1], axes=[dir_axis])',    # h_n
                '!unsqueeze_output(O[2], axes=[dir_axis])',    # c_n
            ),
            attribs={
                'steps': '!I[0].shape[seq_axis]',
                'axis': '!seq_axis',
            },
        ),
})
